"""
Contains a modular preprocessing pipeline in order to generate polar diagrams
from raw data as well as several high level components for said pipeline.
"""

from .expander import Expander, LazyExpander, WeatherExpander
from .extensions import (
    CurveExtension,
    PipelineExtension,
    PointcloudExtension,
    TableExtension,
)
from .pipeline import PipelineOutput, PolarPipeline
from .quality_assurance import (
    ComformingQualityAssurance,
    MinimalQualityAssurance,
    QualityAssurance,
)

__all__ = [
    "PipelineOutput",
    "PolarPipeline",
    "Expander",
    "LazyExpander",
    "WeatherExpander",
    "PipelineExtension",
    "TableExtension",
    "PointcloudExtension",
    "CurveExtension",
    "QualityAssurance",
    "ComformingQualityAssurance",
    "MinimalQualityAssurance",
]



"""
Classes used to transform given data to the hrosailing data format for further processing.

Defines the `DataHandler` abstract base class that can be used to
create custom data handlers to handle formats which are not supported yet.
"""


# pylint: disable=import-outside-toplevel
# pylint: disable=import-error


import csv
from abc import ABC, abstractmethod
from ast import literal_eval
from datetime import date, datetime, time

import numpy as np

from hrosailing.core.constants import HROSAILING_TO_NMEA
from hrosailing.core.data import Data
from hrosailing.core.statistics import ComponentWithStatistics


class DataHandler(ComponentWithStatistics, ABC):
    """Base class for all data handler classes."""

    @abstractmethod
    def handle(self, data):
        """This method should be used to interpret given data in a format
        that is dependent on the handler.

        We recommend using the `hrosailing.pipelinecomponents.data.Data.hrosailing_standard_format`-method
        at the end of your custom handler for compatibility with other built in components.

        Parameters
        ----------
        data :
            Data in a format compatible to the inheriting handler.

        Returns
        -------
        data : Data
            The interpreted data in hrosailing format.
        """

    def set_statistics(self, data):
        """
        Computes standard statistics for the output of a data handler.

        Parameters
        ----------
        data : Data
        """
        super().set_statistics(n_rows=data.n_rows, n_cols=data.n_cols)


class ArrayHandler(DataHandler):
    """A data handler to interpret data given as an array-type.

    See also
    ----------
    `DataHandler`
    """

    # ArrayHandler usable even if pandas is not installed.
    try:
        __import__("pandas")
        pand = True
    except ImportError:
        pand = False

    if pand:
        import pandas as pd

    def handle(self, data):
        """Extracts data from array-types of data.

        Parameters
        ----------
        data : pandas.DataFrame or tuple containing an `array_like` and an ordered iterable
            If given as a tuple, the `array_like` has to contain the values organized in such a way, that the columns
            correspond to different attributes and the rows to different data points.
            In this case, the ordered iterable has to contain the names of the attributes corresponding to the columns.

        See also
        --------
        `DataHandler.handle`
        """
        if self.pand and isinstance(data, self.pd.DataFrame):
            data_dict = data.to_dict(orient="list")
        else:
            arr, keys = data
            arr = np.asarray(arr)

            if len(keys) != arr.shape[1]:
                raise ValueError("Number of keys does not match data")

            data_dict = {key: list(arr[:, i]) for i, key in enumerate(keys)}

        data = Data()
        data.update(data_dict)

        data.hrosailing_standard_format()

        self.set_statistics(data)
        return data


class CsvFileHandler(DataHandler):
    """A data handler to extract data from a .csv file.

    The .csv file has to be ordered in a column-wise fashion, with the
    first entry of each column describing the corresponding attribute.

    Parameters
    ----------
    date_format : str, optional
        Format string compatible with `datetime.strptime` indicating which strings will be treated as dates.

        Defaults to `"%Y-%m-%d %H:%M:%S.%f"`.

    See also
    ----------
    `DataHandler`
    """

    # Check if pandas is available to use from_csv()-method
    try:
        __import__("pandas")
        pand = True
    except ImportError:
        pand = False

    if pand:
        import pandas as pd

    def __init__(self, date_format="%Y-%m-%d %H:%M:%S.%f"):
        super().__init__()
        self._date_format = date_format

    def handle(self, data):
        """
        Parameters
        ----------
        data : path_like
            Path to a .csv file.

        See also
        --------
        `DataHandler.handle`
        """
        if self.pand:
            df = self.pd.read_csv(data)
            data_dict = df.to_dict(orient="list")
        else:
            with open(data, "r", encoding="utf-8") as file:
                csv_reader = csv.reader(file)
                keys = next(csv_reader)
                data_dict = {key: [] for key in keys}
                for row in csv_reader:
                    for i, entry in enumerate(row):
                        data_dict[keys[i]].append(entry)

        for key, value_list in data_dict.items():
            data_dict[key] = [self._save_eval(value) for value in value_list]

        data = Data.from_dict(data_dict)

        data.hrosailing_standard_format()

        self.set_statistics(data)
        return data

    def _save_eval(self, entry):
        if not isinstance(entry, str):
            return entry
        try:
            return datetime.strptime(entry.strip(), self._date_format)
        except (ValueError, TypeError, AttributeError):
            try:
                return literal_eval(entry)
            except SyntaxError as exc:
                raise RuntimeError(
                    f"Could not parse '{entry}'. "
                    "Try using `CsvFileHandler(date_format=...)` "
                    "or check your format string."
                ) from exc


class NMEAFileHandler(DataHandler):
    """A data handler to extract data from a text file containing
    certain NMEA sentences and convert it to a dictionary.

    Parameters
    ----------
    wanted_sentences : iterable of str, optional
        NMEA sentences that will be read.

        Defaults to `None`.

    unwanted_sentences : iterable of str, optional
        NMEA sentences that will be ignored.
        If `wanted_sentences` is set, this keyword is ignored.
        If both `wanted_sentences` and `unwanted_sentences` are `None`,
        all NMEA sentences will be read.

        Defaults to `None`.

    wanted_attributes : iterable of str, optional
        NMEA attributes or hrosailing standard keys that will be appended to the output.

        Defaults to `None`.

    unwanted_attributes : iterable of str, optional
        NMEA attributes that will be ignored.
        If `wanted_attributes` is set, this option is ignored.
        If both `wanted_attributes` and `unwanted_attributes` are `None`,
        all NMEA attributes will be read.

        Defaults to `None`.

    post_filter_types : tuple of types, optional
        The resulting `Data` instance only contains data which is `None` or of a
        type given in `post_filter_types`.
        If set to `False` all attributes are taken into account.

        Defaults to `(float, datetime.date, datetime.time, datetime.datetime)`.

    See also
    ----------
    `DataHandler`
    """

    def __init__(
        self,
        wanted_sentences=None,
        wanted_attributes=None,
        unwanted_sentences=None,
        unwanted_attributes=None,
        post_filter_types=(float, date, time, datetime),
    ):
        super().__init__()
        if wanted_sentences is not None:
            self._sentence_filter = lambda line: any(
                sentence in line for sentence in wanted_sentences
            )
        elif unwanted_sentences is not None:
            self._sentence_filter = lambda line: all(
                sentence not in line for sentence in unwanted_sentences
            )
        else:
            self._sentence_filter = lambda line: True

        if wanted_attributes is not None:
            wanted_nmea_attributes = []
            for key in wanted_attributes:
                if key in HROSAILING_TO_NMEA:
                    wanted_nmea_attributes.extend(HROSAILING_TO_NMEA[key])
                else:
                    wanted_nmea_attributes.append(key)

            self._attribute_filter = lambda field: any(
                field[0] == attribute for attribute in wanted_nmea_attributes
            )
        elif unwanted_attributes is not None:
            self._attribute_filter = lambda field: all(
                field[0] != attribute for attribute in unwanted_attributes
            )
        else:
            self._attribute_filter = lambda field: True

        self._post_filter_types = post_filter_types

    def handle(self, data):
        """Reads a text file containing NMEA sentences and extracts
        data points.

        Parameters
        ----------
        data : path-like
            Path to a text file that contains NMEA 0183 sentences.

        See also
        --------
        `DataHandler.handle`
        """
        from pynmea2 import parse

        comp_data = Data()

        with open(data, "r", encoding="utf-8") as file:
            nmea_sentences = filter(
                self._sentence_filter,
                file,
            )

            for sentence in nmea_sentences:
                parsed_sentence = parse(sentence)
                wanted_fields = filter(
                    self._attribute_filter,
                    parsed_sentence.fields,
                )

                wanted_fields = [x[:2] for x in wanted_fields]

                parsed_dict = {
                    name: getattr(parsed_sentence, attribute)
                    for name, attribute in wanted_fields
                }

                processed_dict = self._postprocess(parsed_dict)

                comp_data.update(processed_dict, compress=True)

        comp_data.hrosailing_standard_format()

        if self._post_filter_types:
            comp_data.filter_types(self._post_filter_types)

        self.set_statistics(comp_data)

        return comp_data

    def _postprocess(self, parsed_sentence):
        if (
            "Wind angle" in parsed_sentence
            and "Reference" in parsed_sentence
            and "Wind speed" in parsed_sentence
        ):
            wind_angle = parsed_sentence["Wind angle"]
            wind_speed = parsed_sentence["Wind speed"]
            reference = parsed_sentence["Reference"]
            if reference == "R":
                parsed_sentence["AWA"] = wind_angle
                parsed_sentence["AWS"] = wind_speed
            elif reference == "T":
                parsed_sentence["TWA"] = wind_angle
                parsed_sentence["TWS"] = wind_speed
            del parsed_sentence["Reference"]
            del parsed_sentence["Wind angle"]
            del parsed_sentence["Wind speed"]
        if "Latitude" in parsed_sentence:
            parsed_sentence["lat"] = self._from_nmea_format(
                parsed_sentence["Latitude"]
            )
            del parsed_sentence["Latitude"]
        if "Longitude" in parsed_sentence:
            parsed_sentence["lon"] = self._from_nmea_format(
                parsed_sentence["Longitude"]
            )
            del parsed_sentence["Longitude"]
        if "Timestamp" in parsed_sentence:
            timestr = parsed_sentence["Timestamp"]
            if isinstance(timestr, str):
                parsed_sentence["Timestamp"] = self._time_from_nmea_format(
                    timestr
                )
        return parsed_sentence

    def _from_nmea_format(self, value):
        value = float(value)
        degrees = int(value / 100)
        minutes = value - 100 * degrees
        return degrees + minutes / 60

    def _time_from_nmea_format(self, timestr):
        hours = int(timestr[0:2])
        minutes = int(timestr[2:4])
        seconds = int(timestr[4:6])
        microseconds = int(timestr.split(".")[1])
        if seconds >= 60:
            minutes += seconds // 60
            seconds %= 60
        if minutes >= 60:
            hours += minutes // 60
            minutes %= 60
        return time(
            hour=hours,
            minute=minutes,
            second=seconds,
            microsecond=microseconds,
        )



# pylint: disable=missing-docstring

import numpy as np


def polynomial_function(x, *params, deg=1):
    """If `deg=n`, models the function
    :math:`a_0 + a_1x + \\dots + a_nx^n`.

    """
    val = 0
    for i in range(deg + 1):
        val += params[i] * np.power(x, i)

    return val


def inverted_parabola(x, *params):
    """Models the function
    :math:`a_1 + a_2(x - a_0)^2`.

    """
    return params[1] + params[2] * np.square(x - params[0])


def inverted_shifted_parabola(x, *params):
    """Models the function
    :math:`a_0x + a_2(x - a_1)^2`.

    Parameters
    -----------

    *params :
        The parameters of the modeled function, i.e.
        :math:`a_0, a_1, a_2`.
    """
    return params[0] * x + params[2] * np.square(x - params[1])


def concave_function(x, *params, downturn=True, sat_limit=False):
    """
    Models a concave function depending on the values of `downturn` and `sat_limits`.
    Note that at least one of the parameters `downturn` and `sat_limit` has to be set to `True`.

    - `downturn` and `sat_limit` are both `True`, models the function
        .. math:: a_2 - a_0e^{a_1 - x} - a_3x^2
    - If `downturn` is `False` and `sat_limit` is `True`, models the function
        .. math:: a_2 - a_0e^{a_1 - x}
    - If `downturn` is `True` and `sat_limits` is `False`, models the function
        .. math:: a_0 + a_1x - a_2x^2

    Parameters
    ----------
    *params :
        The parameters of the modeled function, i.e.
        :math:`a_0, a_1, a_2, a_3`.
        Supported length of `params` depends on the values of `downturn` and `sat_limits`.

    downturn : bool, optional
        Defaults to `True`.

    sat_limit : bool, optional
        Defaults to `False`.
    """
    if sat_limit and downturn:
        return (
            params[2]
            - params[0] * np.exp(params[1] - x)
            - params[3] * np.square(x)
        )

    if sat_limit:
        return params[2] - params[0] * np.exp(params[1] - x)

    if downturn:
        return polynomial_function(x, params[0], params[1], -params[2], deg=2)

    raise ValueError(
        "At least one of `downturn` and `sat_limit` must be `True`"
    )


def s_shaped(x, *params, downturn=True):
    """
    Models the function
    .. math:: \\frac{a_2}{1 + e^{a_0 - a_1x}} - a_3x^2

    Parameters
    -----------

    *params :
        The parameters of the modeled function, i.e.
        :math:`a_0, a_1, a_2, a_3`.
        The supported length depends on the value of `downturn`.

    downturn : bool, optional
        If `False`, we set
        :math:`a_3 = 0`
        independently of the values in `params`.
    """
    val = params[2] / (1 + np.exp(params[0] - params[1] * x))
    if downturn:
        val -= params[3] * np.square(x)

    return val


def gompertz_model(x, *params, neg=False):
    if neg:
        return params[2] * np.exp(-params[0] * np.power(-params[1], x))

    return params[2] * np.exp(-params[0] * np.power(params[1], x))


def gaussian_model(x, *params, offset=False):
    """
    Models the function
    .. math:: a_0e^{-\\frac{(x - a_1)^2}{2a_2}} + a_3

    Parameters
    ----------
    *params :
        The parameters of the modeled function, i.e.
        :math:`a_0, a_1, a_2, a_3`.
        Supported length of `params` depend on the value of `offset`.

    offset : bool, optional
        If `False`, we set
        :math:`a_3 = 0`
        independently of the values in `params`.
    """
    val = params[0] * np.exp(-0.5 * np.square(x - params[1]) / params[2])
    if offset:
        val += params[3]

    return val


def gmm_model(x, *params):
    return (
        params[0] * np.exp(-0.5 * np.square(x - params[1]) / params[2])
        + params[3] * np.exp(-0.5 * np.square(x - params[4]) / params[5])
        + params[6]
    )



# pylint: disable=missing-module-docstring

import csv
from ast import literal_eval
from inspect import getmembers, isfunction

import numpy as np

import hrosailing.core.modelfunctions as model

from ._basepolardiagram import PolarDiagram

MODEL_FUNCTIONS = dict(getmembers(model, isfunction))


class PolarDiagramCurve(PolarDiagram):
    """A class to represent, visualize and work with a polar diagram
    given by a fitted curve/surface.

    Parameters
    ----------
    f : callable
        Curve/surface that describes the polar diagram, given as
        a function, with the signature `f(ws, wa, *params) -> bsp`,
        where

        - `ws` and `wa` should be `array_like` of shape `(n,)`
        and refer to wind speeds and wind angles,
        - `params` is a list of additional parameters,
        - `bsp` is `array_like` of shape `(n,)` and contains the resulting
        boat speeds.

    params : Sequence
        Optimal parameters for `f`.

    radians : bool, optional
        Specifies if `f` takes the wind angles in radians or degrees.

        Defaults to `False`.

    Attributes
    ----------
    curve (property) : function
        A read only version of the underlying function.

    parameters (property) : tuple
        A read only version of the underlying parameters.

    radians (property) : bool
        A read only version wether the wind angle is given in radians or degrees.

    See also
    ----------

    """

    def __init__(self, f, *params, radians=False):
        if not callable(f):
            raise TypeError("`f` is not callable")

        if not self._check_enough_params(f, params):
            raise ValueError(
                "`params` is an incorrect amount of parameters for `f`"
            )

        self._f = f
        self._params = params
        self._rad = radians

    @property
    def default_points(self):
        ws = np.linspace(5, 20, 128)
        wa = np.linspace(5, 355, 144)
        ws, wa = np.meshgrid(ws, wa)
        ws, wa = ws.ravel(), wa.ravel()
        bsp = np.array([self(ws_, wa_) for ws_, wa_ in zip(ws, wa)])
        return np.column_stack([ws, wa, bsp])

    def get_slices(
        self,
        ws=None,
        n_steps=None,
        full_info=False,
        wa_resolution=100,
        **kwargs,
    ):
        """
        Parameters
        ----------------
        wa_resolution : int, optional
            The number of wind angles that will be used for estimation if an
            interpolator is given.

            Defaults to 100.

        See also
        -------
        `PolarDiagram.get_slices`
        """
        kwargs["wa_resolution"] = wa_resolution
        return super().get_slices(ws, n_steps, full_info, **kwargs)

    def ws_to_slices(self, ws, wa_resolution=100, **kwargs):
        """
        See also
        -------
        `PolarDiagramCurve.get_slices`
        `PolarDiagram.ws_to_slices`
        """
        wa_ls = np.linspace(0, 360, wa_resolution)
        return [
            np.row_stack([[ws_] * len(wa_ls), wa_ls, self(ws_, wa_ls)])
            for ws_ in ws
        ]

    @staticmethod
    def _check_enough_params(func, params):
        try:
            func(1, 1, *params)
            return True
        except (IndexError, TypeError):
            return False

    def __repr__(self):
        return (
            f"PolarDiagramCurve(f={self._f.__name__}, "
            f"{str(self._params).strip('()')}, radians={self._rad})"
        )

    def __call__(self, ws, wa):
        # do not change the input wa
        if isinstance(wa, np.ndarray):
            wa = wa.copy()

        if np.any((ws < 0)):
            raise ValueError("`ws` is negative")

        if self.radians:
            wa = np.rad2deg(wa)
            wa %= 360
            wa = np.deg2rad(wa)
        else:
            wa %= 360

        return self.curve(ws, wa, *self.parameters)

    @property
    def curve(self):
        return self._f

    @property
    def default_slices(self):
        return np.linspace(5, 20, 16)

    @property
    def parameters(self):
        """Returns a read only version of `self._params`."""
        return self._params

    @property
    def radians(self):
        """Returns a read only version of `self._rad`."""
        return self._rad

    def to_csv(self, csv_path):
        """Creates a .csv file with delimiter ':' and the
        following format:

            `PolarDiagramCurve`
            Function: `self.curve.__name__`
            Radians: `self.radians`
            Parameters: `self.parameters`

        Parameters
        ----------
        csv_path : path-like
            Path to a .csv file or where a new .csv file will be created.

        Raises
        ------
        OSError
            If no permission to write is given for file.
        """
        with open(csv_path, "w", newline="", encoding="utf-8") as file:
            csv_writer = csv.writer(file, delimiter=":")
            csv_writer.writerow([self.__class__.__name__])
            csv_writer.writerow(["Function"] + [self.curve.__name__])
            csv_writer.writerow(["Radians"] + [str(self.radians)])
            csv_writer.writerow(["Parameters"] + list(self.parameters))

    @classmethod
    def __from_csv__(cls, file):
        csv_reader = csv.reader(file, delimiter=":")
        function = next(csv_reader)[1]
        radians = literal_eval(next(csv_reader)[1])
        params = [literal_eval(value) for value in next(csv_reader)[1:]]

        if function not in MODEL_FUNCTIONS:
            raise RuntimeError(f"no valid function, named {function}")

        function = MODEL_FUNCTIONS[function]
        return PolarDiagramCurve(function, *params, radians=radians)

    def symmetrize(self):
        """Constructs a symmetric version of the polar diagram,
        by mirroring it at the 0° - 180° axis and returning a new instance.
        """

        def sym_func(ws, wa, *params):
            y = self._f(ws, wa, *params)
            y_symm = self._f(ws, (360 - wa), *params)
            return 0.5 * (y + y_symm)

        return PolarDiagramCurve(
            sym_func, *self.parameters, radians=self.radians
        )



# pylint: disable=missing-module-docstring
# pylint: disable=too-many-lines

import csv
import enum
import warnings
from ast import literal_eval
from typing import Iterable

import numpy as np

from hrosailing.core.data import WeightedPoints
from hrosailing.processing import ArithmeticMeanInterpolator, Ball

from ._basepolardiagram import PolarDiagram


class PolarDiagramTable(PolarDiagram):
    """A class to represent, visualize and work with a polar diagram
    in the form of a table.

    Parameters
    ----------
    ws_resolution : array_like, default: `numpy.arange(2, 42, 2)`
        Wind speeds that will correspond to the columns of the table.
        - If array_like, resolution will be `numpy.array(ws_resolution)`.
        - If a scalar `num`, resolution will be `numpy.arange(num, 40, num)`.
    wa_resolution : array_like, default: `numpy.arange(0, 360, 5)`
        Wind angles that will correspond to the rows of the table.
        - If array_like, resolution will be `numpy.array(wa_resolution)`.
        - If a scalar `num`, resolution will be `numpy.arange(num, 360, num)`.
    bsps : array_like, default: `numpy.zeros((rdim, cdim))`
        Boat speeds that will correspond to the entries of the table.
        Needs to have attributes matching `ws_resolution` and `wa_resolution`.

    Attributes
    ----------
    wind_angles (property) : numpy.ndarray
        Read only version of the wind angle resolution.

    wind_speeds (property) : numpy.ndarray
        Read only version of the wind speed resolution.

    boat_speeds (property) : numpy.ndarray
        Read only version of the boat speed table.

    Examples
    --------
    >>> pd = PolarDiagramTable()
    >>> pd.wind_speeds
    [ 2  4  6  8 10 12 14 16 18 20 22 24 26 28 30 32 34 36 38 40]
    >>> pd.wind_angles
    [  0   5  10  15  20  25  30  35  40  45  50  55  60  65  70  75  80
      85  90  95 100 105 110 115 120 125 130 135 140 145 150 155 160 165
     170 175 180 185 190 195 200 205 210 215 220 225 230 235 240 245 250
     255 260 265 270 275 280 285 290 295 300 305 310 315 320 325 330 335
     340 345 350 355]
    >>> pd = PolarDiagramTable(
    ...     bsps=[
    ...         [5.33, 6.32, 6.96, 7.24, 7.35],
    ...         [5.64, 6.61, 7.14, 7.42, 7.56],
    ...         [5.89, 6.82, 7.28, 7.59, 7.84],
    ...         [5.92, 6.98, 7.42, 7.62, 7.93],
    ...         [5.98, 7.07, 7.59, 8.02, 8.34],
    ...         [5.8, 6.95, 7.51, 7.98, 8.52],
    ...         [5.2, 6.41, 7.19, 7.66, 8.14]
    ...     ],
    ...     ws_resolution=[6, 8, 10, 12, 14],
    ...     wa_resolution=[52, 60, 75, 90, 110, 120, 135],
    ... )
    >>> print(pd)
      TWA / TWS    6.0    8.0    10.0    12.0    14.0
    +++++++++++  +++++  +++++  ++++++  ++++++  ++++++
    52.0          5.33   6.32    6.96    7.24    7.35
    60.0          5.64   6.61    7.14    7.42    7.56
    75.0          5.89   6.82    7.28    7.59    7.84
    90.0          5.92   6.98    7.42    7.62    7.93
    110.0         5.98   7.07    7.59    8.02    8.34
    120.0         5.80   6.95    7.51    7.98    8.52
    135.0         5.20   6.41    7.19    7.66    8.14

    Single entries can be extracted like so
    >>> pd[10, 90]
    7.42

    or if a wind speed (resp. wind angle) isn't in the table
    a value can be interpolated
    >>> pd(11, 90)
    7.50924383603392
    """

    def get_slices(
        self,
        ws=None,
        n_steps=None,
        full_info=False,
        interpolator=ArithmeticMeanInterpolator(params=(50,)),
        **kwargs,
    ):
        """
        Parameters
        ----------------
        interpolator : Interpolator, optional
            The interpolator used to estimate boat speeds
            not covered in the table.

            Defaults to `ArithmeticMeanInterpolator(params=(50,)`

        See also
        -------
        `PolarDiagram.get_slices`
        """
        kwargs["interpolator"] = interpolator
        return super().get_slices(ws, n_steps, full_info, **kwargs)

    def ws_to_slices(
        self,
        ws,
        interpolator=ArithmeticMeanInterpolator(params=(50,)),
        **kwargs,
    ):
        """
        See also
        -------
        `PolarDiagramTable.get_slices`
        `PolarDiagram.ws_to_slices`
        """
        slices = []
        for ws_ in ws:
            if ws_ in self.wind_speeds:
                bsp = self.boat_speeds.T[np.where(self.wind_speeds == ws_)][0]
            else:
                bsp = [
                    self(ws_, wa_, interpolator) for wa_ in self.wind_angles
                ]
            slices.append(
                np.row_stack(
                    [[ws_] * len(self.wind_angles), self.wind_angles, bsp]
                )
            )
        return slices

    def __init__(self, ws_resolution=None, wa_resolution=None, bsps=None):
        ws_resolution = _Resolution.WIND_SPEED.set_resolution(ws_resolution)
        wa_resolution = _Resolution.WIND_ANGLE.set_resolution(wa_resolution)

        if bsps is None:
            self._create_zero_table(ws_resolution, wa_resolution)
            return

        bsps = np.asarray_chkfinite(bsps, float)

        # sanity checks
        if bsps.dtype is object:
            raise TypeError("`bsps` is not array_like")
        if _incompatible_shapes(bsps, ws_resolution, wa_resolution):
            raise ValueError("`bsps` has incorrect shape")

        (
            self._ws_resolution,
            self._wa_resolution,
            self._boat_speeds,
        ) = _sort_table(ws_resolution, wa_resolution, bsps)

    def _create_zero_table(self, ws_resolution, wa_resolution):
        rows, cols = len(wa_resolution), len(ws_resolution)
        self._boat_speeds = np.zeros((rows, cols))
        self._ws_resolution = sorted(ws_resolution)
        self._wa_resolution = sorted(wa_resolution)

    def __str__(self):
        table = ["  TWA / TWS"]
        ws = self.wind_speeds
        if len(ws) <= 15:
            self._create_short_table(table, ws)
        else:
            wind = []
            wind.extend(ws[:5])
            wind.extend(ws[-5:])
            self._create_long_table(table, wind)

        return "".join(table)

    def _create_short_table(self, table, wind):
        bsps = self.boat_speeds

        table.extend([f"    {float(ws):.1f}" for ws in wind])
        table.append("\n+++++++++++")

        for ws in wind:
            le = len(f"{float(ws):.1f}")
            table.append("  ".ljust(le + 4, "+"))

        table.append("\n")

        for i, wa in enumerate(self.wind_angles):
            table.append(f"{float(wa):.1f}".ljust(11))

            for j, ws in enumerate(wind):
                le = len(f"{float(ws):.1f}")
                table.append(f"{bsps[i][j]:.2f}".rjust(4 + le))

            table.append("\n")

    def _create_long_table(self, table, wind):
        bsps = self.boat_speeds

        for i, ws in enumerate(wind):
            if i == 5:
                table.append("  ...")

            table.append(f"    {float(ws):.1f}")

        table.append("\n+++++++++++")

        for i, ws in enumerate(wind):
            if i == 5:
                table.append("  +++")

            le = len(f"{float(ws):.1f}")
            table.append("  ".ljust(le + 4, "+"))

        table.append("\n")

        for i, wa in enumerate(self.wind_angles):
            table.append(f"{float(wa):.1f}".rjust(11))

            for j, ws in enumerate(wind):
                if j == 5:
                    table.append("  ...")

                le = len(f"{float(ws):.1f}")
                table.append(f"{bsps[i][j]:.2f}".rjust(4 + le))

            table.append("\n")

    def __repr__(self):
        return (
            f"PolarDiagramTable({self.wind_speeds}, {self.wind_angles}, "
            f"{self.boat_speeds})"
        )

    def __call__(
        self,
        ws,
        wa,
        interpolator=ArithmeticMeanInterpolator(params=(50,)),
        neighbourhood=Ball(radius=1),
    ):
        """Calculates the boat speed for given `ws` and `wa`.

        If the `ws-wa` point is in the table, the corresponding entry is
        returned, otherwise the value is interpolated.

        Parameters
        ----------
        ws : int or float
            Wind speed.

        wa : int or float
            Wind angle.

        interpolator : hrosailing.pipelinecomponents.Interpolator, optional
            Interpolator subclass that determines the interpolation
            method used to determine the value at the ws-wa point.

            Defaults to: `hrosailing.pipelinecomponents.ArithmeticMeanInterpolator(params = (50, ))`

        neighbourhood : hrosailing.pipelinecomponents.Neighbourhood, optional
            Neighbourhood subclass used to determine the grid points
            in the table that will be used in the interpolation.

            Defaults to `hrosailing.pipelinecomponents.Ball(1)`

        Returns
        -------
        bsp : int or float
            Boat speed value as determined above.
        """
        try:
            return self[ws, wa]
        except (TypeError, ValueError):
            point = np.array([ws, wa])
            ws, wa = np.meshgrid(self.wind_speeds, self.wind_angles)
            points = np.column_stack(
                (ws.ravel(), wa.ravel(), self.boat_speeds.ravel())
            )

            weighted_points = WeightedPoints(points, weights=1)

            considered_points = neighbourhood.is_contained_in(
                points[:, :2] - point
            )

            return interpolator.interpolate(
                weighted_points[considered_points], point
            )

    def __getitem__(self, *key):
        """Returns the value of a given entry in the table."""
        ws, wa = key[0]

        col = self._get_indices(np.atleast_1d(ws), _Resolution.WIND_SPEED)
        row = self._get_indices(np.atleast_1d(wa), _Resolution.WIND_ANGLE)
        return float(self.boat_speeds[row, col])

    def _get_indices(self, wind, soa):
        res = (
            self.wind_speeds
            if soa.name == _Resolution.WIND_SPEED.name
            else self.wind_angles
        )

        if wind is None:
            return range(len(res))

        wind = soa.normalize_wind(wind)

        # sanity checks
        if not wind:
            raise ValueError("empty slice-list was passed")

        if not wind.issubset(set(res)):
            raise ValueError(f"{wind} is not contained in {res}")

        return [i for i, w in enumerate(res) if w in wind]

    @property
    def wind_angles(self):
        """Returns a read only version of `self._wa_resolution`."""
        return self._wa_resolution.copy()

    @property
    def wind_speeds(self):
        """Returns a read only version of `self._ws_resolution`."""
        return self._ws_resolution.copy()

    @property
    def default_slices(self):
        """
        See also
        ---------
        `Polardiagram.default_slices`
        """
        return self.wind_speeds

    @property
    def default_points(self):
        """
        See also
        --------
        `Polardiagram.default_slices`
        """
        x, y = np.meshgrid(self.wind_speeds, self.wind_angles)
        wind = np.array(list(zip(x.ravel(), y.ravel())))
        bsps = self.boat_speeds.ravel()
        return np.column_stack([wind, bsps])

    @property
    def boat_speeds(self):
        """Returns a read only version of `self._boat_speeds`."""
        return self._boat_speeds.copy()

    def to_csv(self, csv_path, fmt="hro"):
        """Creates a .csv file with delimiter ',' and the following format:

            `PolarDiagramTable`
            TWS:
            `self.wind_speeds`
            TWA:
            `self.wind_angles`
            Boat speeds:
            `self.boat_speeds`

        Parameters
        ----------
        csv_path : path_like
            Path for the .csv file.
        fmt : {"hro", "orc", "opencpn", "array"}, optional
            Format for the .csv file.

            Defaults to "hro".

        See Also
        --------
        `from_csv`

        Examples
        --------
        >>> pd = PolarDiagramTable(
        ...     bsps=[
        ...         [5.33, 6.32, 6.96, 7.24, 7.35],
        ...         [5.64, 6.61, 7.14, 7.42, 7.56],
        ...         [5.89, 6.82, 7.28, 7.59, 7.84],
        ...         [5.92, 6.98, 7.42, 7.62, 7.93],
        ...         [5.98, 7.07, 7.59, 8.02, 8.34],
        ...         [5.8, 6.95, 7.51, 7.98, 8.52],
        ...         [5.2, 6.41, 7.19, 7.66, 8.14]
        ...     ],
        ...     ws_resolution=[6, 8, 10, 12, 14],
        ...     wa_resolution=[52, 60, 75, 90, 110, 120, 135],
        ... )
        >>> print(pd)
          TWA / TWS    6.0    8.0    10.0    12.0    14.0
        +++++++++++  +++++  +++++  ++++++  ++++++  ++++++
        52.0          5.33   6.32    6.96    7.24    7.35
        60.0          5.64   6.61    7.14    7.42    7.56
        75.0          5.89   6.82    7.28    7.59    7.84
        90.0          5.92   6.98    7.42    7.62    7.93
        110.0         5.98   7.07    7.59    8.02    8.34
        120.0         5.80   6.95    7.51    7.98    8.52
        135.0         5.20   6.41    7.19    7.66    8.14
        >>> pd.to_csv("example.csv")
        >>> pd2 = from_csv("example.csv")
        >>> print(pd2)
          TWA / TWS    6.0    8.0    10.0    12.0    14.0
        +++++++++++  +++++  +++++  ++++++  ++++++  ++++++
        52.0          5.33   6.32    6.96    7.24    7.35
        60.0          5.64   6.61    7.14    7.42    7.56
        75.0          5.89   6.82    7.28    7.59    7.84
        90.0          5.92   6.98    7.42    7.62    7.93
        110.0         5.98   7.07    7.59    8.02    8.34
        120.0         5.80   6.95    7.51    7.98    8.52
        135.0         5.20   6.41    7.19    7.66    8.14
        """
        if fmt not in {"hro", "orc", "opencpn", "array"}:
            raise NotImplementedError("`fmt` not implemented")

        with open(csv_path, "w", newline="", encoding="utf-8") as file:
            if fmt == "orc":
                self._write_orc_format(file)
                return

            if fmt == "array":
                self._write_array_format(file)
                return

            csv_writer = csv.writer(file, delimiter=",")

            if fmt == "opencpn":
                self._write_opencpn_format(csv_writer)
                return

            self._write_hro_format(csv_writer)

    def _write_orc_format(self, file):
        csv_writer = csv.writer(file, delimiter=";")
        csv_writer.writerow(["twa/tws"] + list(self.wind_speeds))
        csv_writer.writerow([0] * (len(self.wind_speeds) + 1))

        self._write_rows(csv_writer)

    def _write_rows(self, csv_writer):
        rows = np.column_stack((self.wind_angles, self.boat_speeds))
        csv_writer.writerows(rows)

    def _write_array_format(self, file):
        csv_writer = csv.writer(file, delimiter="\t")
        csv_writer.writerow([r"TWA\TWS"] + list(self.wind_speeds))

        self._write_rows(csv_writer)

    def _write_opencpn_format(self, csv_writer):
        csv_writer.writerow([r"TWA \ TWS"] + list(self.wind_speeds))

        self._write_rows(csv_writer)

    def _write_hro_format(self, csv_writer):
        csv_writer.writerow([self.__class__.__name__])
        csv_writer.writerow(["TWS"])
        csv_writer.writerow(self.wind_speeds)
        csv_writer.writerow(["TWA"])
        csv_writer.writerow(self.wind_angles)
        csv_writer.writerow(["BSP"])
        csv_writer.writerows(self.boat_speeds)

    @classmethod
    def __from_csv__(cls, file):
        csv_reader = csv.reader(file, delimiter=",")
        next(csv_reader)
        ws_res = [literal_eval(ws) for ws in next(csv_reader)]
        next(csv_reader)
        wa_res = [literal_eval(wa) for wa in next(csv_reader)]
        next(csv_reader)
        bsps = [[literal_eval(bsp) for bsp in row] for row in csv_reader]

        return PolarDiagramTable(ws_res, wa_res, bsps)

    def symmetrize(self):
        """Constructs a symmetric version of the polar diagram.

        Symmetrizes polar diagram by mirroring it at the 0° - 180° axis
        and returning a new instance.

        Warnings
        --------
        Should only be used if all the wind angles of the initial
        polar diagram are on one side of the 0° - 180° axis,
        otherwise this can lead to duplicate data, which can
        overwrite or live alongside old data.

        Examples
        --------
        >>> pd = PolarDiagramTable(
        ...     bsps=[
        ...         [5.33, 6.32, 6.96, 7.24, 7.35],
        ...         [5.64, 6.61, 7.14, 7.42, 7.56],
        ...         [5.89, 6.82, 7.28, 7.59, 7.84],
        ...         [5.92, 6.98, 7.42, 7.62, 7.93],
        ...     ],
        ...     ws_resolution = [6, 8, 10, 12, 14],
        ...     wa_resolution = [52, 60, 75, 90]
        ... )
        >>> print(pd)
          TWA / TWS    6.0    8.0    10.0    12.0    14.0
        +++++++++++  +++++  +++++  ++++++  ++++++  ++++++
        52.0          5.33   6.32    6.96    7.24    7.35
        60.0          5.64   6.61    7.14    7.42    7.56
        75.0          5.89   6.82    7.28    7.59    7.84
        90.0          5.92   6.98    7.42    7.62    7.93
        >>> sym_pd = pd.symmetrize()
        >>> print(sym_pd)
          TWA / TWS    6.0    8.0    10.0    12.0    14.0
        +++++++++++  +++++  +++++  ++++++  ++++++  ++++++
        52.0          5.33   6.32    6.96    7.24    7.35
        60.0          5.64   6.61    7.14    7.42    7.56
        75.0          5.89   6.82    7.28    7.59    7.84
        90.0          5.92   6.98    7.42    7.62    7.93
        270.0         5.92   6.98    7.42    7.62    7.93
        285.0         5.89   6.82    7.28    7.59    7.84
        300.0         5.64   6.61    7.14    7.42    7.56
        308.0         5.33   6.32    6.96    7.24    7.35
        """

        symmetric_wa_resolution = np.concatenate(
            [self.wind_angles, 360 - np.flip(self.wind_angles)]
        )
        symmetric_bsps = np.row_stack(
            (self.boat_speeds, np.flip(self.boat_speeds, axis=0))
        )

        if 180 in self.wind_angles:
            (
                symmetric_wa_resolution,
                symmetric_bsps,
            ) = _delete_multiple_180_degree_occurrences(
                symmetric_wa_resolution, symmetric_bsps
            )
        if 0 in self.wind_angles:
            (
                symmetric_wa_resolution,
                symmetric_bsps,
            ) = _delete_multiple_0_degree_occurrences(
                symmetric_wa_resolution, symmetric_bsps
            )

        return PolarDiagramTable(
            ws_resolution=self.wind_speeds,
            wa_resolution=symmetric_wa_resolution,
            bsps=symmetric_bsps,
        )

    def change_entries(self, new_bsps, ws=None, wa=None):
        """Changes specified entries in the table.

        Parameters
        ----------
        new_bsps: array_like of matching shape
            Sequence containing the new boat speeds to be inserted
            in the specified entries.
        ws: array_like, optional
            Columns of table given as elements of `self.wind_speed`.

            Defaults to `self.wind_speeds`.

        wa: array_like, optional
            Rows of table given as elements of `self.wind_angles`.

            Defaults to `self.wind_angles`.

        Examples
        --------
        >>> pd = PolarDiagramTable(
        ...     ws_resolution=[6, 8, 10, 12, 14],
        ...     wa_resolution=[52, 60, 75, 90, 110, 120, 135]
        ... )
        >>> print(pd)
          TWA / TWS    6.0    8.0    10.0    12.0    14.0
        +++++++++++  +++++  +++++  ++++++  ++++++  ++++++
        52.0          0.00   0.00    0.00    0.00    0.00
        60.0          0.00   0.00    0.00    0.00    0.00
        75.0          0.00   0.00    0.00    0.00    0.00
        90.0          0.00   0.00    0.00    0.00    0.00
        110.0         0.00   0.00    0.00    0.00    0.00
        120.0         0.00   0.00    0.00    0.00    0.00
        135.0         0.00   0.00    0.00    0.00    0.00
        >>> pd.change_entries(
        ...     new_bsps=[5.33, 5.64, 5.89, 5.92, 5.98, 5.8, 5.2],
        ...     ws=6
        ... )
        >>> print(pd)
          TWA / TWS    6.0    8.0    10.0    12.0    14.0
        +++++++++++  +++++  +++++  ++++++  ++++++  ++++++
        52.0          5.33   0.00    0.00    0.00    0.00
        60.0          5.64   0.00    0.00    0.00    0.00
        75.0          5.89   0.00    0.00    0.00    0.00
        90.0          5.92   0.00    0.00    0.00    0.00
        110.0         5.98   0.00    0.00    0.00    0.00
        120.0         5.80   0.00    0.00    0.00    0.00
        135.0         5.20   0.00    0.00    0.00    0.00
        >>> pd.change_entries(
        ...     new_bsps=[5.7, 6.32, 6.96, 7.24, 7.35],
        ...     wa=52
        ... )
        >>> print(pd)
          TWA / TWS    6.0    8.0    10.0    12.0    14.0
        +++++++++++  +++++  +++++  ++++++  ++++++  ++++++
        52.0          5.70   6.32    6.96    7.24    7.35
        60.0          5.64   0.00    0.00    0.00    0.00
        75.0          5.89   0.00    0.00    0.00    0.00
        90.0          5.92   0.00    0.00    0.00    0.00
        110.0         5.98   0.00    0.00    0.00    0.00
        120.0         5.80   0.00    0.00    0.00    0.00
        135.0         5.20   0.00    0.00    0.00    0.00
        """
        new_bsps = np.atleast_1d(new_bsps)

        new_bsps = np.asarray_chkfinite(new_bsps)

        if new_bsps.dtype == object:
            raise TypeError("`new_bsps` is not array_like")

        ws = self._get_indices(ws, _Resolution.WIND_SPEED)
        wa = self._get_indices(wa, _Resolution.WIND_ANGLE)

        wa_len = len(wa) == 1
        ws_len = len(ws) == 1

        # wrong shape can lead to missing assignments
        if wa_len and ws_len:
            correct_shape = new_bsps.shape == (1,)
        elif wa_len:
            correct_shape = new_bsps.shape == (len(ws),)
        elif ws_len:
            correct_shape = new_bsps.shape == (len(wa),)
        else:
            correct_shape = new_bsps.shape == (len(wa), len(ws))

        if not correct_shape:
            raise ValueError("`new_bsps` has wrong shape")

        mask = np.zeros(self.boat_speeds.shape, dtype=bool)
        for i in wa:
            for j in ws:
                mask[i, j] = True

        self._boat_speeds[mask] = new_bsps.flat


class _Resolution(enum.Enum):
    WIND_SPEED = (np.arange(2, 42, 2), 40)
    WIND_ANGLE = (np.arange(0, 360, 5), 360)

    def __init__(self, standard_res, max_value):
        self.standard_res = standard_res
        self.max_value = max_value

    def set_resolution(self, res):
        if res is None:
            return self.standard_res

        if isinstance(res, Iterable):
            return self._custom_iterable_resolution(res)

        return self._custom_stepsize_resolution(res)

    def _custom_iterable_resolution(self, res):
        # NaN's and infinite values can cause problems later on
        res = np.asarray_chkfinite(res)

        if res.dtype == object:
            raise ValueError("`res` is not array_like")

        if not res.size or res.ndim != 1:
            raise ValueError("`res` has incorrect shape")

        if len(set(res)) != len(res):
            warnings.warn(
                "`res` contains duplicate data. "
                "This may lead to unwanted behaviour"
            )

        if self.name == _Resolution.WIND_SPEED.name:
            if np.any((res < 0)):
                raise ValueError("`res` contains negative entries")

        if self.name == _Resolution.WIND_ANGLE.name:
            res %= 360

        return res

    def _custom_stepsize_resolution(self, res):
        if res <= 0:
            raise ValueError("`res` is non-positive")

        if res > self.max_value:
            raise ValueError(
                f"Resolution stepsize ({res}) is bigger than the maximal"
                f" resolution value ({self.max_value})."
            )

        return np.arange(res, self.max_value, res)

    def normalize_wind(self, wind):
        wind = np.atleast_1d(wind)  # allow scalar inputs

        if self.name == _Resolution.WIND_SPEED.name:
            if np.any((wind < 0)):
                raise ValueError("`wind` is negative")

        if self.name == _Resolution.WIND_ANGLE.name:
            wind %= 360

        return set(wind)


def _incompatible_shapes(bsps, ws_resolution, wa_resolution):
    rows, cols = len(wa_resolution), len(ws_resolution)
    return bsps.shape != (rows, cols)


def _sort_table(ws_resolution, wa_resolution, bsps):
    wa_resolution, bsps = zip(
        *sorted(zip(wa_resolution, bsps), key=lambda x: x[0])
    )
    bsps = np.array(bsps, float)

    ws_resolution, bsps = zip(
        *sorted(zip(ws_resolution, bsps.T), key=lambda x: x[0])
    )

    return (
        np.array(ws_resolution, float),
        np.array(wa_resolution, float),
        np.array(bsps, float).T,
    )


def _delete_multiple_180_degree_occurrences(wa_resolution, bsps):
    mid = np.where(wa_resolution == 180)[0][0]
    wa_resolution = np.delete(wa_resolution, mid)
    bsps = np.row_stack((bsps[:mid, :], bsps[mid + 1 :, :]))

    return wa_resolution, bsps


def _delete_multiple_0_degree_occurrences(wa_resolution, bsps):
    return wa_resolution[:-1], bsps[:-1, :]



"""
Contains the baseclass `Smoother` used in the `PolarPipeline` class
that can also be used to create custom smoothers.

Also contains predefined and usable smoothers:

- `LazySmoother`,
- `AffineSmoother`.
"""
import datetime
from abc import ABC, abstractmethod

from hrosailing.core.statistics import ComponentWithStatistics


class Smoother(ComponentWithStatistics, ABC):
    """
    Base class for all smoothers aimed to smoothen measurement errors in given data.
    """

    @abstractmethod
    def smooth(self, data):
        """
        Should be used to smoothen the measurement errors in `data`, where `data`
        is interpreted as a time series.

        Parameters
        ----------
        data : Data
            The data to be smoothened.

        Returns
        -------
        data : Data
            The processed data.
        """


class LazySmoother(Smoother):
    """
    Smoother that doesn't do anything.

    See also
    ----------
    `Smoother`
    """

    def smooth(self, data):
        """
        Does not change the data.

        See also
        --------
        `Smoother.smooth`
        """
        return data


class AffineSmoother(Smoother):
    """
    Locates intervals in which the data, interpreted as a time series,
    is constant.
    Between two intervals, the arithmetic mean of the values of the respective
    intervals is used to replace the measured data.
    Then, data points in such an interval are replaced using affine
    interpolation.

    Parameters
    ---------
    timespan: datetime.timedelta, optional
        Amount of time before the actual value of an interval is assumed to be the real value.

        Defaults to `timedelta(seconds=30)`.

    See also
    ----------
    `Smoother`
    """

    def __init__(self, timespan=datetime.timedelta(seconds=30)):
        super().__init__()
        self._timespan = timespan.total_seconds()

    def smooth(self, data):
        """
        Smooths data using the procedure described above.

        See also
        --------
        `Smoother.smooth`
        """
        if any(time < data["datetime"][0] for time in data["datetime"]):
            raise ValueError(
                "AffineSmoother only supports chronologically ordered time"
                " series."
            )
        for key in data.keys():
            if data.type(key) == float:
                data = self._smooth_field(key, data)
        return data

    def _smooth_field(self, key, data):
        ys = data[key].copy()
        start_time = data["datetime"][0]
        xs = [(time - start_time).total_seconds() for time in data["datetime"]]

        interval_bounds_x, interval_bounds_y = self._get_interval_bounds(
            xs, ys
        )

        approx_intervals_x, approx_intervals_y = self._approximate_intervals(
            interval_bounds_x, interval_bounds_y, ys
        )

        smooth_data = self._smooth_data_from_approx_intervals(
            approx_intervals_x, approx_intervals_y, xs, data, key
        )

        return smooth_data

    def _get_interval_bounds(self, xs, ys):
        x_lb = xs[0]
        y_lb = ys[0]
        interval_bounds_x = []
        interval_bounds_y = []
        for i, (x, x_after, y, y_after) in enumerate(
            zip(xs, xs[1:], ys, ys[1:])
        ):
            if i == len(ys) - 2:
                interval_bounds_x.append((x_lb, x_after))
                interval_bounds_y.append(y)
                break
            if y == y_after:
                continue
            interval_bounds_x.extend([(x_lb, 1 / 2 * (x + x_after))])
            interval_bounds_y.append(y_lb)
            x_lb = 1 / 2 * (x + x_after)
            y_lb = y_after

        return interval_bounds_x, interval_bounds_y

    def _approximate_intervals(self, interval_bounds_x, interval_bounds_y, ys):
        approx_intervals_x = []
        approx_intervals_y = []

        y_before = ys[0]
        # for (x_lb, x_ub), y, y_after in zip(
        #         interval_bounds_x, interval_bounds_y, interval_bounds_y[1:] + [ys[-1]]
        # ):
        for (x_lb, x_ub), y, y_after in zip(
            interval_bounds_x,
            interval_bounds_y,
            interval_bounds_y[1:] + [ys[-1]],
        ):
            if x_ub - x_lb < 2 * self._timespan:
                midpt = 1 / 2 * (x_ub + x_lb)
                approx_intervals_x.extend([(x_lb, midpt), (midpt, x_ub)])
                approx_intervals_y.extend(
                    [(1 / 2 * (y_before + y), y), (y, 1 / 2 * (y_after + y))]
                )
            else:
                lb_ref = x_lb + self._timespan
                ub_ref = x_ub - self._timespan
                approx_intervals_x.extend(
                    [(x_lb, lb_ref), (lb_ref, ub_ref), (ub_ref, x_ub)]
                )
                approx_intervals_y.extend(
                    [
                        (1 / 2 * (y_before + y), y),
                        (y, y),
                        (y, 1 / 2 * (y_after + y)),
                    ]
                )
            y_before = y

        return approx_intervals_x, approx_intervals_y

    def _smooth_data_from_approx_intervals(
        self, approx_intervals_x, approx_intervals_y, xs, data, key
    ):
        x_lb, x_ub = approx_intervals_x[0]
        y_lb, y_ub = approx_intervals_y[0]
        interval_idx = 0
        for i, x in enumerate(xs):
            try:
                while x > x_ub:
                    interval_idx += 1
                    x_lb, x_ub = approx_intervals_x[interval_idx]
                    y_lb, y_ub = approx_intervals_y[interval_idx]
            except IndexError:
                break

            try:
                mu = (x - x_lb) / (x_ub - x_lb)
                data[key][i] = mu * y_ub + (1 - mu) * y_lb
            except ZeroDivisionError:
                data[key][i] = y_ub

        return data



"""
Contains constants used in the pipeline.

- `KEYSYNONYMS` is a dictionary mapping commonly used sailing terms
(in lower case and the words are " " separated)
to the hrosailing standard term,

- `SEPARATORS` is a list containing characters that will be interpreted as separators between words when applying the
hrosailing standard format,

- `NORM_SCALES` is a dictionary containing suggested scale factors for various hrosailing standard keywords. These scales
are intended to be used to scale components before applying a norm,

- `HROSAILING_TO_NMEA` is a dictionary containing `key`, `value` pairs where `key` is a hrosailing standard keyword and
`value` is a list of NMEA 0183 keywords which are required for the computation of the data associated with `key`.
"""

KEYSYNONYMS = {
    "wind": "TWS",
    "wind speed": "TWS",
    "true": "TWS",
    "true wind": "TWS",
    "true wind speed": "TWS",
    "ws": "TWS",
    "tws": "TWS",
    "wspd": "TWS",
    "apparent wind speed": "AWS",
    "apparent wind": "AWS",
    "apparent": "AWS",
    "aws": "AWS",
    "angle": "TWA",
    "wind angle": "TWA",
    "true angle": "TWA",
    "true wind angle": "TWA",
    "wa": "TWA",
    "twa": "TWA",
    "apparent wind angle": "AWA",
    "apparent angle": "AWA",
    "awa": "AWA",
    "speed": "BSP",
    "boat speed": "BSP",
    "boat speed knots": "BSP",
    "vessel speed": "BSP",
    "vessel speed knots": "BSP",
    "sp": "BSP",
    "bsp": "BSP",
    "bsps": "BSP",
    "water speed": "BSP",
    "water speed knots": "BSP",
    "wsp": "BSP",
    "speed over ground": "SOG",
    "speed over ground knots": "SOG",
    "og": "SOG",
    "sog": "SOG",
    "spd over grnd": "SOG",
    "date": "date",
    "datestamp": "date",
    "date stamp": "date",
    "time": "time",
    "timestamp": "time",
    "time stamp": "time",
    "datetime": "datetime",
    "date time": "datetime",
    "lat": "lat",
    "latitude": "lat",
    "lon": "lon",
    "long": "lon",
    "longitude": "lon",
    "temp": "temp",
    "temperature": "temp",
    "dewpoint": "dewpoint",
    "dewpt": "dewpoint",
    "dwpt": "dewpoint",
    "dwpoint": "dewpoint",
    "rel humidity": "humidity",
    "relative humidity": "humidity",
    "rel hum": "humidity",
    "relative hum": "humidity",
    "rhum": "humidity",
    "relhum": "humidity",
    "hum": "humidity",
    "humidity": "humidity",
    "wd": "WD",
    "wdir": "WD",
    "winddir": "WD",
    "wpgt": "gust",
    "pres": "air pressure",
    "tsun": "total sunshine",
}


HROSAILING_TO_NMEA = {
    "TWS": ["Wind speed", "Reference"],
    "AWS": ["Wind speed", "Reference"],
    "TWA": ["Wind angle", "Reference"],
    "AWA": ["Wind angle", "Reference"],
    "lat": ["Latitude"],
    "lon": ["Longitude"],
    "SOG": ["Speed Over Ground"],
    "date": ["Datestamp"],
    "time": ["Timestamp"],
}

SEPARATORS = [",", ";", "_", "-", "+", "&", "\n", "\t"]

NORM_SCALES = {
    "TWS": 1 / 20,
    "AWS": 1 / 20,
    "TWA": 1 / 360,
    "AWA": 1 / 360,
    "BSP": 1 / 40,
    "SOG": 1 / 40,
    "lat": 1 / 360,
    "lon": 1 / 360,
}



"""
`PolarDiagram` classes to work with and represent polar diagrams in various
forms.
"""

from ._basepolardiagram import PolarDiagram
from ._polardiagramcurve import PolarDiagramCurve
from ._polardiagrammultisails import PolarDiagramMultiSails
from ._polardiagrampointcloud import PolarDiagramPointcloud
from ._polardiagramtable import PolarDiagramTable
from ._reading import from_csv

__all__ = [
    "PolarDiagram",
    "PolarDiagramCurve",
    "PolarDiagramMultiSails",
    "PolarDiagramPointcloud",
    "PolarDiagramTable",
    "from_csv",
]



"""
Classes used for modular modelling of different regression methods.

Defines the `Regressor` abstract base class that can be used to create
custom regression methods.

Subclasses of `Regressor` can be used with the `CurveExtension` class
in the `hrosailing.pipeline` module.
"""


import inspect
from abc import ABC, abstractmethod
from typing import Callable

import numpy as np
from scipy.odr.odrpack import ODR, Data, Model
from scipy.optimize import curve_fit


class Regressor(ABC):
    """Base class for all regressor classes."""

    @property
    @abstractmethod
    def model_func(self):
        """
        A version of the model function used in the regression.
        """

    @property
    @abstractmethod
    def optimal_params(self):
        """
        A version of the optimal parameters determined
        through regression of the model function.
        """

    @abstractmethod
    def fit(self, data):
        """This method should, given data, be used to determine
        optimal parameters for the model function.

        Parameters
        ----------
        data : array_like of shape (n, 3)
            Data to which the model function will be fitted, given as
            a sequence of points consisting of wind speed, wind angle
            and boat speed.
        """


class ODRegressor(Regressor):
    """An orthogonal distance regressor based on `scipy.odr.odrpack`.

    Parameters
    ----------
    model_func : function
        The function to be fitted.

        The function signature has to be `f(ws, wa, *params) -> bsp`,
        where `ws` and `wa` are `numpy.ndarrays` resp. and `params` is a
        sequence of parameters that will be fitted.

    init_values : array_like
        Initial guesses for the optimal parameters of `model_func`
        that are passed to the `scipy.odr.ODR` class.

    max_it : int, optional
        Maximum number of iterations done by `scipy.odr.ODR`.

        Defaults to `1000`.

    See also
    ----------
    `Regressor`
    """

    def __init__(self, model_func: Callable, init_values, max_it=1000):
        def odr_model_func(params, x):
            ws = x[0, :]
            wa = x[1, :]
            return model_func(ws, wa, *params)

        self._func = model_func
        self._model = Model(odr_model_func)
        self._init_vals = init_values
        # self._weights_X = None
        # self._weights_y = None
        self._maxit = max_it
        self._popt = None

    @property
    def model_func(self):
        """Returns a read-only version of `self._func`.

        See also
        ----------
        `Regressor.model_func`"""
        return self._func

    @property
    def optimal_params(self):
        """Returns a read-only version of `self._popt`.

        See also
        ----------
        `Regressor.optimal_params`"""
        return self._popt

    def fit(self, data):
        """Fits the model function to the given data, i.e. calculates
        the optimal parameters to minimize an objective
        function based on the given data.

        Parameters
        ----------
        data : array_like of shape (n, 3)
            Data to which the model function will be fitted, given as
            a sequence of points consisting of wind speed, wind angle
            and boat speed.

        See also
        --------
        [ODRPACK](https://docs.scipy.org/doc/external/odrpack_guide.pdf), `Regressor.fit`
        """
        if data.shape[1] != 3:
            raise ValueError("`data` has to be of shape (n, 3).")
        if data.shape[0] == 0:
            raise ValueError("`data` has to contain at least one record!")
        X, y = data[:, :2], data[:, 2]

        odr_data = Data((X[:, 0], X[:, 1]), y)
        odr = ODR(
            odr_data, self._model, beta0=self._init_vals, maxit=self._maxit
        )
        odr.set_job(fit_type=2)
        out = odr.run()

        self._popt = out.beta


class LeastSquareRegressor(Regressor):
    """A least square regressor based on `scipy.optimize.curve_fit`.

    Parameters
    ----------
    model_func : function or callable
        The function which describes the model and is to be fitted.

        The function signature has to be `f(ws, wa, *params) -> bsp`,
        where `ws` and `wa` are `numpy.ndarrays` resp. and `params` is a
        sequence of parameters that will be fitted.

    init_vals : array_like, optional
        Initial guesses for the optimal parameters of `model_func`
        that are passed to `scipy.optimize.curve_fit`.

        Defaults to `None`.

    See also
    ----------
    `Regressor`
    """

    def __init__(self, model_func: Callable, init_vals=None):
        self._func = model_func

        def fitting_func(wind, *params):
            wind = np.asarray(wind)
            ws, wa = wind[:, 0], wind[:, 1]
            return model_func(ws, wa, *params)

        sig = inspect.signature(model_func)
        args = [
            p.name
            for p in sig.parameters.values()
            if p.kind
            in [
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                inspect.Parameter.POSITIONAL_ONLY,
            ]
        ]

        if init_vals is None and len(args) < 3:
            init_vals = _determine_params(fitting_func)
        elif init_vals is None:
            init_vals = tuple(1 for _ in range(len(args)))

        self._fitting_func = fitting_func
        self._init_vals = init_vals
        self._popt = None

    @property
    def model_func(self):
        """Read-only version of `self._func`.

        See also
        ----------
        `Regressor.model_func`
        """
        return self._func

    @property
    def optimal_params(self):
        """Read-only version of `self._popt`.

        See also
        ----------
        `Regressor.optimal_params`
        """
        return self._popt

    def fit(self, data):
        """Fits the model function to the given data, i.e. calculates
        the optimal parameters to minimize the sum of the squares of
        the residuals.

        Parameters
        ----------
        data : array_like of shape (n, 3)
            Data to which the model function will be fitted, given as
            a sequence of points consisting of wind speed, wind angle
            and boat speed.

        See also
        --------
        [curve_fit](https://docs.scipy.org/doc/scipy/reference/generated/\
        scipy.optimize.curve_fit.html), `Regressor.fit`
        """
        X, y = data[:, :2], data[:, 2]

        self._popt = self._get_optimal_parameters(X, y)

        # if _enable_logging:
        #     self._log_outcome_of_regression(X, y)

    def _get_optimal_parameters(self, X, y):
        optimal_parameters, _ = curve_fit(
            self._fitting_func, X, y, p0=self._init_vals
        )
        return optimal_parameters


def _determine_params(func):
    params = []
    while True:
        try:
            func(np.array([[0, 0]]), *params)
            break
        except (IndexError, TypeError):
            params.append(1)

    return params



"""
Functions for basic computations.
"""


import enum

import numpy as np


class _Wind(enum.Enum):
    TO_TRUE = -1
    TO_APPARENT = 1

    def convert_wind(self, wind):
        """
        converts true wind to apparent and vice versa
        """
        wind = np.asarray_chkfinite(wind)

        if wind.dtype == object:
            raise TypeError("`wind` is not array_like")
        if wind.ndim != 2 or wind.shape[1] != 3:
            raise TypeError("`wind` has incorrect shape")

        ws, wa, bsp = np.hsplit(wind, 3)
        if np.any((ws < 0)):
            raise TypeError("`wind` has negative wind speeds")

        wa %= 360  # normalize wind angles
        wa_above_180 = wa > 180
        wa = np.deg2rad(wa)

        converted_ws = self._convert_wind_speed(ws, wa, bsp)
        converted_wa = self._convert_wind_angle(
            converted_ws, ws, wa, bsp, wa_above_180
        )

        return np.column_stack((converted_ws, converted_wa, bsp))

    def _convert_wind_speed(self, ws, wa, bsp):
        return np.sqrt(
            np.square(ws)
            + np.square(bsp)
            + 2 * self.value * ws * bsp * np.cos(wa)
        )

    def _convert_wind_angle(self, converted_ws, ws, wa, bsp, wa_above_180):
        temp = (ws * np.cos(wa) + self.value * bsp) / converted_ws

        # account for floating point errors
        temp[temp > 1] = 1
        temp[temp < -1] = -1

        converted_wa = np.arccos(temp)
        converted_wa[wa_above_180] = 360 - np.rad2deg(
            converted_wa[wa_above_180]
        )
        converted_wa[~wa_above_180] = np.rad2deg(converted_wa[~wa_above_180])

        return converted_wa


def convert_apparent_wind_to_true(apparent_wind):
    """Convert apparent wind to true wind.

    Parameters
    ----------
    apparent_wind : array_like of shape (n, 3)
        Wind data given as a sequence of points consisting of wind speed,
        wind angle and boat speed, where the wind speed and wind angle are
        measured as apparent wind.

    Returns
    -------
    true_wind : numpy.ndarray of shape (n, 3)
        Array containing the same data as `apparent_wind`, but the wind speed
        and wind angle now measured as true wind.
    """
    return _Wind.TO_TRUE.convert_wind(apparent_wind)


def convert_true_wind_to_apparent(true_wind):
    """Convert true wind to apparent wind.

    Parameters
    ----------
    true_wind : array_like of shape (n, 3)
        Wind data given as a sequence of points consisting of wind speed,
        wind angle and boat speed, where the wind speed and wind angle are
        measured as true wind.

    Returns
    -------
    apparent_wind : numpy.ndarray of shape (n, 3)
        Array containing the same data as `true_wind`, but the wind speed
        and wind angle now measured as apparent wind.
    """
    return _Wind.TO_APPARENT.convert_wind(true_wind)


def scaled_norm(norm, scal_factors):
    """
    Returns a function acting as a scaled norm.

    Parameters
    ----------
    norm : function
        A function mapping `numpy.ndarray` (vector(s)) to `float` (the norm of
        the vector).

    scal_factors : array-like of floats
        Component-wise scaling factors. Shape has to be compatible with the
        possible input shapes of `norm`.

    Returns
    ---------
    s_norm : `norm` scaled component-wise by `scal_factors`.
    """
    scaled = np.asarray(scal_factors)

    def s_norm(vec):
        return norm(scaled * vec)

    return s_norm


def euclidean_norm(vec):
    """
    Evaluates the euclidean norm on a two-dimensional array.

    Parameters
    --------
    vec : numpy.ndarray of shape (n, d)

    Returns
    -------
    norm : numpy.ndarray of shape (n)
        The euclidean norms of the columns of `vec`.
    """
    return np.linalg.norm(vec, axis=1)


def scaled_euclidean_norm(vec):
    """
    Scaled version of the euclidean norm. Scaling factors depend on the input
    dimension. First component gets scaled by 1/40, second by 1/360 and third
    by 1/20.

    Parameters
    --------
    vec : numpy.ndarray of shape (n, d)

    Returns
    -------
    norm : numpy.ndarray of shape (n)
        The scaled euclidean norms of the columns of `vec`.
    """
    if vec.shape[1] == 2:
        norm_val = scaled_norm(euclidean_norm, [1 / 40, 1 / 360])(vec)
    elif vec.shape[1] == 3:
        norm_val = scaled_norm(euclidean_norm, [1 / 40, 1 / 360, 1 / 20])(vec)
    else:
        raise NotImplementedError(
            "scaled_euclidean_norm only supports 2 and 3 dimensional inputs"
        )
    return norm_val


def data_dict_to_numpy(data_dict, keys):
    """
    Method to transform a dictionary of lists of floats into a numpy array.

    Parameter
    ---------
    data_dict : dict,
        Dictionary to transform.

    keys : [str],
        Keys that indicate which lists of the data dictionary will be used
        to create the columns of the resulting array.

    Returns
    --------
    (n, d) array where 'n' is the length of a list in the data dictionary and
    'd' is 'len(keys)'.
    """
    return np.column_stack([data_dict[key] for key in keys])


def safe_operation(operand, value):
    """
    Perform an operation savely, returning `None` if one of the following
    Exceptions is raised:
    `ValueError`, `TypeError`, `IndexError`, `KeyError`, `ZeroDivisionError`.

    Parameters
    ---------
    operand : function
        The operation to be savely evaluated.

    value : arbitrary
        The input value of the operation.
    """
    try:
        return operand(value)
    except (ValueError, TypeError, IndexError, KeyError, ZeroDivisionError):
        return None



"""Contains the class `Data` which is an output of several pipeline components.
"""

import itertools
from datetime import datetime
from decimal import Decimal

import numpy as np

from .constants import KEYSYNONYMS, SEPARATORS


class Data:
    """
    Organizes measurement data. Data is interpreted as a table with columns
    identified by keywords and rows corresponding to a certain measurement
    point. Also saves and checks the data types of the corresponding columns.
    Weights might be added to the corresponding rows.

    If `data` is of type `Data`, `data[key]` yields a reduced version of `data` depending on the type of `key`:

     - if `key` is a string it returns the corresponding column.
     - if `key` is an integer it returns the corresponding row.
     - if `key` is an array like of strings, it works identical to `data.get_slice(key)`
     - if `key` is an array like of booleans, it uses `key` as a boolean mask on the rows.

    `key in data` checks if there is a column with key `key`.

    Iteration is supported and is performed over the rows.

    Also supports the `str` method.
    """

    def __init__(self):
        self._data = {}
        self._types = {}
        self._max_len = 0

    def keys(self):
        """
        Returns
        -------
        keys : list of str
            List of all column names.
        """
        return list(self._data.keys())

    @property
    def numerical(self):
        """
        A `numpy.ndarray` containing all data of type `float`.
        """
        float_keys, float_vals = self.get_by_type(float)
        try:
            array = np.column_stack(float_vals)
        except ValueError:
            array = np.empty((0,))
        return float_keys, array

    @property
    def n_rows(self):
        """
        The number of rows.
        """
        return self._max_len

    @property
    def n_cols(self):
        """
        The number of columns.
        """
        return len(self._data)

    @property
    def data(self):
        """
        A 'read-only' version of the dictionary of columns.
        """
        return self._data.copy()

    def rows(self, keys=None, return_type=dict):
        """
        Iterates over the rows.

        Parameters
        ----------
        keys : iterable over str, optional
            The columns from which the iterator yields. Other columns will be ignored.

            Defaults to `self.keys()`.

        return_type : dict, tuple or list
            Defines the type of the yielded data. If types `tuple` or `list` are used, the entries are in the exact same
            order as in `keys`.

            Defaults to `dict`.

        Yields
        -------
        The columns of the instance, filtered as indicated by `keys` organized
        in a way indicated by `return_type`.
        """
        if keys is None:
            keys = self.keys()
        for i in range(self._max_len):
            if return_type is dict:
                yield {key: self._data[key][i] for key in keys}
            elif return_type is tuple:
                yield (self._data[key][i] for key in keys)
            elif return_type is list:
                yield [self._data[key][i] for key in keys]
            else:
                raise ValueError(
                    f"Return type of {return_type} is not supported"
                )

    def get_by_type(self, type_):
        """
        Parameters
        ----------
        type_ : type

        Returns
        -------
        filtered_keys : list of str
            A list of all keys with columns of type `type_`.

        filtered_data : list of lists
            A list of all data corresponding to the columns with keys
            `filtered_keys`.
        """
        filtered_keys = [
            key for key in self.keys() if self._types[key] == type_
        ]
        return filtered_keys, [self._data[key] for key in filtered_keys]

    def type(self, key):
        """
        Parameters
        ----------
        key : str

        Returns
        --------
        type_ : type
            The associated type of the column corresponding to `key`.
        """
        return self._types[key]

    def extend(self, key, data):
        """
        Extends the data by given other data.

        Parameters
        ----------
        key : str
            The key of the column that will be extended by the given data.
        data : list
            Data that will be appended to the given column.
        """
        data_type = self._get_type(data)
        if key in self._data:
            if self._types[key] is None:
                self._types[key] = data_type
            if data_type is not None and data_type != self._types[key]:
                raise ValueError(
                    f"data should be of type {self._types[key]}"
                    f" but has type {data_type}"
                )
            self._data[key].extend(data)
            self._max_len = max(self._max_len, len(self._data[key]))
        else:
            # self.fill(self._max_len - len(data), [key])
            # self._data[key].extend(data)
            self._data[key] = data
            self._types[key] = data_type
            self._max_len = max(self._max_len, len(self._data[key]))

    def update(self, data_dict, compress=False):
        """
        Extends the data according to given data and fills missing
        entries in each column with `None`.

        Parameters
        ----------
        data_dict : dict or Data
            The dictionary or `Data` object containing the data to be used for the update.

        compress : bool, optional
            If `True` and the last entrys of all present keys are not set or
            `None`, the last entries will be replaced.
        """
        if isinstance(data_dict, Data):
            self.update(data_dict.data)
            return
        if not isinstance(data_dict, dict):
            raise TypeError("`data_dict` has to be of type `Data` or `dict`")
        self.fill(
            self._max_len, itertools.chain(data_dict.keys(), self.keys())
        )
        if not compress or self._max_len == 0:
            self._unsafe_update(data_dict)
            return

        if all(self[key][-1] is None for key in data_dict.keys()):
            for key in data_dict.keys():
                self._data[key] = self._data[key][:-1]

        self._unsafe_update(data_dict)

    def _unsafe_update(self, data_dict):
        for key, val in data_dict.items():
            if isinstance(val, list):
                self.extend(key, val)
            else:
                self.append(key, val)
        self.fill()

    def append(self, key, data):
        """
        Extends the data by a single element.

        Parameters
        ----------
        key : str
            The key of the column to which the given data will be appended.
        data : object
            Single data object to be appended.

        See also
        --------
        `Data.extend`
        """
        self.extend(key, [data])

    def fill(self, len_=None, keys=None):
        """
        Fills all columns smaller than a specific length with `None` values.

        Parameters
        ----------
        len_ : int, optional
            The required length.

            Defaults to `self.n_rows`.

        keys : list of str, optional
            The keys of the columns which will be filled.

            Defaults to `self.keys`.
        """
        if keys is None:
            keys = self.keys()
        if len_ is None:
            len_ = self._max_len
        for key in keys:
            if key not in self._data:
                self._data[key] = []
            curr_len = len(self._data[key])
            fill_len = max(0, (len_ - curr_len))
            self._data[key].extend([None] * fill_len)
            self._types[key] = self._get_type(self._data[key])

    def filter_types(self, type_list):
        """
        Deletes all data which is not in a list of required types.

        Parameters
        ----------
        type_list : list of types
        """
        for key in list(self._data.keys()):
            if self._types[key] not in type_list:
                del self._data[key]
                del self._types[key]

    def rename(self, old_key, new_key):
        """
        Renames a column.

        Parameters
        ----------
        old_key : str
            Old name of column.

        new_key : str
            New name of column.
        """
        if old_key == new_key:
            return
        if new_key in self.keys():
            raise ValueError(
                f"Can not rename {old_key} to {new_key} since"
                f" {new_key} already exists"
            )

        self._data[new_key] = self._data[old_key]
        self._types[new_key] = self._types[old_key]

        self.delete(old_key)

    def delete(self, key):
        """
        Deletes a column or a row.

        Parameters
        ----------
        key : int, str, list of str or list of int
            Name(s) of column(s) to be deleted or index(es) of row(s) to be deleted.
        """
        if isinstance(key, str):
            del self._data[key]
            del self._types[key]
        if isinstance(key, int):
            self.delete([key])
        if isinstance(key, list):
            if len(key) == 0:
                return
            if isinstance(key[0], str):
                for k in key:
                    self.delete(k)
            if isinstance(key[0], int):
                self._data = {
                    k: [v for i, v in enumerate(value) if i not in key]
                    for k, value in self._data.items()
                }
                self._max_len -= len(key)

    def strip(self, mode="all"):
        """
        Removes either columns that only consist of `None` values or leading and tailing rows consisting only of `None`
        values.

        Parameters
        ----------
        mode : {"cols", "rows", "all"}
            Specifies whether columns or rows or both will be removed.

            Defaults to `"all"`.
        """
        if mode == "cols":
            self._data = {
                key: value
                for key, value in self._data.items()
                if not all(v is None for v in value)
            }
        elif mode == "rows":
            end_leading_nones = 0  # first index AFTER leading None rows
            start_tailing_nones = (
                self._max_len - 1
            )  # first index BEFORE tailing None rows
            while all(
                value[end_leading_nones] is None
                for key, value in self._data.items()
            ):
                end_leading_nones += 1
            while all(
                value[start_tailing_nones] is None
                for key, value in self._data.items()
            ):
                start_tailing_nones -= 1
            self._data = {
                key: value[end_leading_nones : start_tailing_nones + 1]
                for key, value in self._data.items()
            }
        elif mode == "all":
            self.strip("cols")
            self.strip("rows")

    def hrosailing_standard_format(self):
        """
        Reformats data into the `hrosailing` standard format.

        This means:
            - the dictionary has `hrosailing` standard keys whenever possible,
            - date and time fields will be aggregated to `datetime`,
            - tries to cast entries to `float` whenever possible.
        """

        def get_standard_key(key):
            lkey = key.lower()
            for sep in SEPARATORS:
                lkey = lkey.replace(sep, " ")
            lkey = lkey.strip()
            # remove plural
            stripped_lkey = lkey if lkey[-1] != "s" else lkey[:-1]
            if lkey in KEYSYNONYMS:
                return KEYSYNONYMS[lkey]
            if stripped_lkey in KEYSYNONYMS:
                return KEYSYNONYMS[stripped_lkey]

            return key

        for key, value in list(self._data.items()):
            standard_key = get_standard_key(key)
            new_key = standard_key
            if key == new_key:
                continue
            i = 0
            while new_key in self.keys():
                i += 1
                new_key = f"{standard_key}(Control value {i})"
            self.rename(key, new_key)

        if "time" in self and "date" in self:

            def combine(date, time):
                if date is None or time is None:
                    return None
                return datetime.combine(date, time)

            self.extend(
                "datetime",
                [
                    combine(date, time)
                    for date, time in zip(self["date"], self["time"])
                ],
            )
            self.delete("date")
            self.delete("time")

        # ensure floats
        for key in self.keys():
            if self._types[key] in (int, str, Decimal, np.float_):
                succes, self._data[key] = _try_call_to_float(self._data[key])
                if succes:
                    self._types[key] = float

    @staticmethod
    def _get_type(data):
        curr_type = None
        for entry in data:
            if entry is None:
                continue
            if curr_type is None:
                curr_type = type(entry)
                continue
            if type(entry) != curr_type:
                raise ValueError(
                    "Data has no consistent type."
                    f"Found the types {type(entry)} and {curr_type}"
                )
        return curr_type

    @classmethod
    def concatenate(cls, list_):
        """
        Returns concatenated data, given a list of `Data` instances.

        Parameters
        ----------
        list_ : list of Data

        Returns
        -------
        data : Data
            The concatenated Data.
        """
        data = cls()
        for other_data in list_:
            data.update(other_data)
        return data

    @classmethod
    def _force_set(cls, data, types, max_len):
        # use with caution: wrong usage can produce inconsistent `Data` objects.
        new_obj = cls()
        new_obj._data = data
        new_obj._types = types
        new_obj._max_len = max_len
        return new_obj

    @classmethod
    def from_dict(cls, dict_):
        """
        Creates a `Data` object that contains the same data as a given dictionary.
        The keys and values of the dictionary need to be iterables and will correspond to the columns of the resulting
        `Data` object.

        Parameters
        ----------
        dict_ : dict

        Returns
        -------
        data : Data
            Instance containing the same data as `dict_`.
        """
        data = cls()
        data.update(dict_)
        return data

    def get_slice(self, slice_):
        """
        Returns new `Data` object containing only keys in `slice_`.

        Parameters
        ----------
        slice_ : object with method `__contains__`

        Returns
        -------
        data: Data
        """

        data = {
            key: value for key, value in self._data.items() if key in slice_
        }
        types = {
            key: value for key, value in self._types.items() if key in slice_
        }
        try:
            max_len = max(len(field) for field in data.values())
        except ValueError:
            max_len = 0
        return Data._force_set(data, types, max_len)

    def _mask_rows(self, mask):
        """
        Keeps only a subset of the rows indicated by `mask`.

        Parameters
        ----------
        mask : array_like of booleans
        """

        data = {
            key: [entry for entry, choose in zip(value, mask) if choose]
            for key, value in self._data.items()
        }
        try:
            max_len = max(len(field) for field in data.values())
        except ValueError:
            max_len = 0

        return Data._force_set(data, self._types.copy(), max_len)

    def __getitem__(self, item):
        if isinstance(item, str):
            return self._data[item]
        if isinstance(item, int):
            return {key: val[item] for key, val in self._data.items()}
        type_ = None
        for key in item:
            type_ = type(key)
            break
        if type_ is None or type_ in (bool, np.bool_):
            return self._mask_rows(item)
        if type_ is str:
            return self.get_slice(item)
        raise TypeError(
            "Only types `int`, `str` and iterables over `bool`, `numpy.bool_`"
            " or `str` are supported"
        )

    def __contains__(self, item):
        return item in self._data

    def __iter__(self):
        for i in range(self._max_len):
            yield self[i]

    def __str__(self):
        str_ = ""
        for key in self.keys():
            str_ += f"Data field '{key}' of type {self._types[key]}:\n\t"
            if len(self._data[key]) < 10:
                str_ += str(self._data[key])
            else:
                str_ += (
                    f"[{','.join([str(item) for item in self._data[key][:5]])},"
                    " ..."
                    f" ,{','.join([str(item) for item in self._data[key][-5:]])}]"
                )
            str_ += "\n"

        return str_


def _try_call_to_float(list_):
    new_list = []
    for i, value in enumerate(list_):
        if value is None:
            new_list.append(value)
            continue
        try:
            new_list.append(float(value))
        except (TypeError, ValueError):
            return False, list_
    return True, new_list


class WeightedPoints:
    """A class to weigh data points and represent them together
    with their respective weights.

    If `weighted_points` is of type `WeightedPoints` then you can use `weighted_points[mask]` with an array-like over
    booleans to create a boolean mask over the respective rows of the weighted points.

    Parameters
    ----------
    data : Data, dict or numpy.ndarray
        Points that will be weight or paired with given weights.
        If given as a dictionary, each value has to be a list. Data points are interpreted as
        `[data[key][i] for key in data.keys()]` for each suitable `i`.
        If given as a `numpy.ndarray`, the rows will be interpreted as data points.

    weights : scalar or array_like of shape (n, )
        If the weights of the points are known beforehand,
        they can be given as an argument. If weights are
        passed, they will be assigned to the points
        and no further weighing will take place.

        If a scalar is passed, the points will all be assigned
        the same weight.
    """

    def __init__(self, data, weights):
        self.data = data
        if isinstance(weights, (float, int)):
            if isinstance(data, dict):
                length = len(list(data.values())[0])
            else:
                length = len(data)
            self.weights = weights * np.ones(length)
        else:
            self.weights = np.asarray(weights)

    def __getitem__(self, mask):
        if isinstance(self.data, dict):
            return WeightedPoints(
                data={
                    key: list(np.array(value)[mask])
                    for key, value in self.data.items()
                },
                weights=self.weights[mask],
            )
        return WeightedPoints(data=self.data[mask], weights=self.weights[mask])

    def extend(self, other):
        """
        Extends the weighted points by other weighted points.
        The value of the `data` attribute has to be of the same type in both respective `WeightedPoints` objects.
        If both data is given as a dictionary of lists, the respective lists
        will be extended.
        Keys that are not present in both dictionaries are discarded.

        Parameters
        ----------
        other : WeightedPoints
            Points to be appended.
        """

        if isinstance(self.data, dict):
            self.data = {
                key: value + other.data[key]
                for key, value in self.data.items()
                if key in other.data
            }
        elif isinstance(self.data, Data):
            self.data = Data.concatenate([self.data, other.data])
        else:
            self.data = np.row_stack([self.data, other.data])

        self.weights = np.concatenate([self.weights, other.weights])



"""
Contains projections and functions for plotting objects of the hrosailing framework.
Currently, the plot of `PolarDiagram` objects is supported.
Defines the following projections:

- `"hro polar"` : plot and scatter polar diagrams in a polar plot (see `HROPolar`),
- `"hro flat"` : plot and scatter polar diagrams in a euclidean plot (see `HROFlat`),
- `"hro color gradient"` : plot two-dimensional heat maps of a polar diagram (see `HROColorGradient`),
- `"hro 3d"` : scatter or plot the surface of the three-dimensional representation of a polar diagram (see `Axes3D`).

Examples
--------
>>> from polardiagram._reading import from_csv import matplotlib.pyplot as plt
>>> import hrosailing.plotting
>>>
>>> ax = plt.subplot("hro polar")
>>> pd = from_csv("my_file.pd")
>>> ax.plot(pd)
>>> plt.show()
"""

import itertools

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from matplotlib.cm import ScalarMappable
from matplotlib.colors import (
    LinearSegmentedColormap,
    Normalize,
    is_color_like,
    to_rgb,
)
from matplotlib.lines import Line2D
from matplotlib.projections import register_projection
from matplotlib.projections.polar import PolarAxes
from matplotlib.tri import Triangulation
from mpl_toolkits.mplot3d import Axes3D as pltAxes3D
from scipy.spatial import ConvexHull
from scipy.spatial.qhull import QhullError

from hrosailing.polardiagram import PolarDiagram


class HROPolar(PolarAxes):
    """
    Projection for plotting polar diagrams in a polar plot.
    """

    name = "hro polar"

    def plot(
        self,
        *args,
        ws=None,
        n_steps=None,
        colors=("green", "red"),
        show_legend=False,
        legend_kw=None,
        use_convex_hull=False,
        **kwargs,
    ):
        """
        Plots the given data as a polar plot.
        If a `PolarDiagram` is given, plots each slice corresponding to `ws` and `n_steps`
        as described in `PolarDiagram.get_slices`.

        Parameters
        ----------
        *args :
            If the first argument is a polar diagram it plots the polar diagram.
            Otherwise, it plots the arguments as usual.

        ws : int, float or iterable thereof, optional
            As in `Polardiagram.get_slices`.

        n_steps : int, optional
            As in `Polardiagram.get_slices`.

        colors : color_like or sequence of color_likes or (ws, color_like) pairs
            Specifies the colors to be used for the different slices.

            - If a color_like is passed, all slices will be plotted in the
            respective color.
            - If 2 colors are passed, slices will be plotted with a color
            gradient that is determined by the corresponding wind speed.
            - Otherwise the slices will be colored in turn with the specified
            colors or the color `"blue"`, if there are too few colors. The
            order is determined by the corresponding wind speeds.
            - Alternatively one can specify certain slices to be plotted in
            a color out of order by passing a sequence of `(ws, color)` pairs.

            Defaults to `("green", "red")`.

        show_legend : bool
            Specifies whether a legend will be shown next to the plot.

            The type of legend depends on the color options.

            If plotted with a color gradient, a `matplotlib.colorbar.Colorbar`
            will be created, otherwise a `matplotlib.legend.Legend` instance.

            Defaults to `False`.

        legend_kw : dict, optional
            Keyword arguments to change position and appearance of the colorbar
            or legend respectively.
            - If 2 colors are passed, a colorbar will be created.
            In this case see `matplotlib.colorbar.Colorbar` for possible
            keywords and their effect.
            - Otherwise, a legend will be created.
            In this case see `matplotlib.legend.Legend` for possible keywords
            and their effect.

            Defaults to `None`

        use_convex_hull : bool, optional
            If set to `True`, the convex hull (in polar coordinates) of the slices will be plotted instead of the
            slices itself.
        """
        if not isinstance(args[0], PolarDiagram):
            super().plot(*args, **kwargs)
            return

        labels, slices, info = self._prepare_plot(
            args, ws, n_steps, colors, show_legend, legend_kw, **kwargs
        )
        _plot(self, slices, info, True, use_convex_hull, **kwargs)

    def scatter(
        self,
        *args,
        ws=None,
        n_steps=None,
        colors=("green", "red"),
        show_legend=False,
        legend_kw=None,
        use_convex_hull=False,
        **kwargs,
    ):
        """
        Plots the given data as a polar plot.
        Parameters are identical to `plot`.

        See also
        -------
        `HROPolar.plot`
        """
        if not isinstance(args[0], PolarDiagram):
            super().scatter(*args, **kwargs)
            return

        labels, slices, info = self._prepare_plot(
            args, ws, n_steps, colors, show_legend, legend_kw, **kwargs
        )
        _plot(self, slices, info, True, use_convex_hull, True, **kwargs)

    def _prepare_plot(
        self, args, ws, n_steps, colors, show_legend, legend_kw, **kwargs
    ):
        pd = args[0]
        labels, slices, info = pd.get_slices(ws, n_steps, full_info=True)
        _set_polar_axis(self)
        _configure_axes(self, labels, colors, show_legend, legend_kw, **kwargs)
        return labels, slices, info


class HROFlat(Axes):
    """
    Projection to plot given data in a rectilinear plot.
    """

    name = "hro flat"

    def plot(
        self,
        *args,
        ws=None,
        n_steps=None,
        colors=("green", "red"),
        show_legend=False,
        legend_kw=None,
        use_convex_hull=False,
        **kwargs,
    ):
        """
        Plots the given data in a rectilinear plot.
        Otherwise, it works identical to `HROPolar.plot`.

        See also
        ----------
        `HROPolar.plot`
        """
        if not isinstance(args[0], PolarDiagram):
            super().plot(*args, **kwargs)
            return

        pd = args[0]
        labels, slices, info = pd.get_slices(ws, n_steps, full_info=True)
        _configure_axes(self, labels, colors, show_legend, legend_kw, **kwargs)
        _plot(self, slices, info, False, use_convex_hull, **kwargs)

    def scatter(
        self,
        *args,
        ws=None,
        n_steps=None,
        colors=("green", "red"),
        show_legend=False,
        legend_kw=None,
        use_convex_hull=False,
        **kwargs,
    ):
        """
        Creates rectilinear scatter plot of a `PolarDiagram` instance.
        Otherwise, it works the same as `HROPolar.plot`

        See also:
        ----------
        `HROPolar.plot`
        """

        if not isinstance(args[0], PolarDiagram):
            super().scatter(*args, **kwargs)
            return

        pd = args[0]
        labels, slices, info = pd.get_slices(ws, n_steps, full_info=True)
        _configure_axes(self, labels, colors, show_legend, legend_kw, **kwargs)
        _plot(self, slices, info, False, use_convex_hull, True, **kwargs)


class HROColorGradient(Axes):
    """Projection supporting two-dimensional color gradient plotting of polar diagrams."""

    name = "hro color gradient"

    def plot(self, *args, **kwargs):
        """
        Works identical to `scatter`.

        See also
        ----------
        `HROColorGradient.scatter`
        """
        self.scatter(*args, **kwargs)

    def scatter(
        self,
        *args,
        wind=None,
        colors=("green", "red"),
        show_legend=False,
        legend_kw=None,
        **kwargs,
    ):
        """
        Plots the given data as a polar plot.
        If a `PolarDiagram` is given, plots each slice corresponding to `ws` and `n_steps`
        as described in `PolarDiagram.get_slices`.

        Parameters
        ----------
        *args :
            If the first argument is a polar diagram it plots the polar diagram.
            Otherwise, it plots the arguments as usual.

        wind : (2, d) or (d, 2) numpy.ndarray or tuple of 2 (1, d) numpy.ndarray or `None`, optional
            As in `Polardiagram.get_points`.

        colors : color_like or sequence of color_likes or (ws, color_like) pairs
            Specifies the colors to be used for the different slices.
            As in `HROPolar.plot`.

        show_legend : bool, default: `False`
            Specifies whether a legend will be shown next to the plot.
            As in `HROPolar.plot`.

        legend_kw : dict, optional
            Keyword arguments to change position and appearance of the colorbar
            or legend respectively.
            As in `HROPolar.plot`.
        """
        if not isinstance(args[0], PolarDiagram):
            super().scatter(*args, **kwargs)
            return

        if legend_kw is None:
            legend_kw = {}

        pd = args[0]
        points = pd.get_points(wind=wind)
        ws, wa, bsp = points.T

        if show_legend:
            _configure_legend(self, bsp, colors, "Boat Speed", **legend_kw)

        color_gradient = _determine_color_gradient(colors, bsp.ravel())

        self.scatter(ws, wa, c=color_gradient, **kwargs)


class Axes3D(pltAxes3D):
    """Projection enabling the display of polar diagrams in a three-dimensional plot."""

    name = "hro 3d"

    def scatter(
        self,
        *args,
        wind=None,
        colors=("green", "red"),
        **kwargs,
    ):
        """
        Works identical to `HROColorGradient.scatter`.

        See also
        ---------
        `HROColorGradient.scatter`
        """
        if not isinstance(args[0], PolarDiagram):
            super().scatter(*args, **kwargs)
            return

        pd = args[0]
        x, y, z = self._prepare_points(pd, wind)

        self._plot3d(x, y, z, colors, **kwargs)

    def plot(self, *args, **kwargs):
        """
        Works identical to `plot_surface` on polar diagrams.
        Otherwise, it works like `mpl_toolkits.mplot3d.Axes3D.plot`.

        See also
        --------
        `Axes3D.plot_surface`
        `mpl_toolkits.mplot3d.Axes3D.plot`
        """
        if not isinstance(args[0], PolarDiagram):
            super().plot(*args, **kwargs)
            return

        self.plot_surface(*args, **kwargs)

    def plot_surface(
        self, *args, wind=None, colors=("green", "red"), **kwargs
    ):
        """
        Plots a three-dimensional depiction of the polar diagram as a triangulated surface plot if a
        polar diagram is given. Otherwise, it works as `mpl_toolkits.mplot3d.plot_surface`.
        Parameters are the same as in `HROColorGradient.scatter`.

        See also
        --------
        `HROColorGradient.scatter`
        `mpl_toolkits.mplot3d.Axes3D.plot_surface`
        """
        if not isinstance(args[0], PolarDiagram):
            super().plot_surface(*args, **kwargs)
            return

        pd = args[0]
        x, y, z = self._prepare_points(pd, wind)

        triang = Triangulation(x, y)

        txs = x[triang.triangles]
        tys = y[triang.triangles]
        tzs = z[triang.triangles]

        diffz = [
            (tzs[:, idx1] - tzs[:, idx2]) ** 2
            for idx1, idx2 in itertools.combinations([0, 1, 2], 2)
        ]

        not_too_narrow = np.logical_or(
            diffz[0] > 0.2, diffz[1] > 0.2, diffz[2] > 0.2
        )
        axis_skip = np.logical_or(
            np.sign(txs[:, 0]) != np.sign(txs[:, 1]),
            np.sign(txs[:, 0]) != np.sign(txs[:, 2]),
        )
        northern = np.logical_and(*[tys[:, i] > 0 for i in range(2)])
        no_northern_skip = np.logical_not(np.logical_and(axis_skip, northern))
        mask = np.logical_and(not_too_narrow, no_northern_skip)

        color_map = _create_color_map(colors)

        _set_3d_axis_labels(self)
        _remove_3d_tick_labels_for_polar_coordinates(self)

        self.plot_trisurf(
            x, y, z, triangles=triang.triangles[mask], cmap=color_map, **kwargs
        )

    def _prepare_points(self, pd, wind):
        points = pd.get_points(wind)
        ws, wa, bsp = points.T
        # flip and rotate such that 0° is on top and 90° is right
        # wa = (-wa)%360 - 90
        wa_rad = np.deg2rad(wa)

        x, y = bsp * np.sin(wa_rad), bsp * np.cos(wa_rad)
        return x, y, ws

    def _plot3d(self, x, y, z, colors, **plot_kw):
        _set_3d_axis_labels(self)
        _remove_3d_tick_labels_for_polar_coordinates(self)

        color_map = _create_color_map(colors)

        super().scatter(x, y, z, c=z, cmap=color_map, **plot_kw)


register_projection(HROPolar)
register_projection(HROFlat)
register_projection(HROColorGradient)
register_projection(Axes3D)


def _plot(
    ax,
    slices,
    info,
    use_radians,
    use_convex_hull=False,
    use_scatter=False,
    **kwargs,
):
    def safe_zip(iter1, iter2):
        if iter2 is not None:
            yield from zip(iter1, iter2)
            return
        for entry in iter1:
            yield (entry, None)

    if len(slices) == 0:
        ax.plot([], [], **kwargs)
        return

    for slice_, info_ in safe_zip(slices, info):
        slice_ = slice_[:, np.argsort(slice_[1])]
        if use_convex_hull:
            ws, wa, bsp, info_ = _get_convex_hull(slice_, info_)
        else:
            ws, wa, bsp = slice_
        if use_radians:
            wa = np.deg2rad(wa)
        if info_ is not None and not use_convex_hull:
            wa, bsp = _alter_with_info(wa, bsp, info_)
        if use_scatter:
            ax.scatter(wa, bsp, **kwargs)
        else:
            ax.plot(wa, bsp, **kwargs)


def _get_info_intervals(info_):
    intervals = {}
    for j, entry in enumerate(info_):
        if entry not in intervals:
            intervals[entry] = []
        intervals[entry].append(j)
    return list(intervals.values())


def _merge(wa, intervals):
    if len(intervals) == 0:
        return np.empty((0))
    wa_in_intervals = [
        np.concatenate([wa[interval], [np.NAN]]) for interval in intervals
    ]
    return np.concatenate(wa_in_intervals)[:-1]


def _alter_with_info(wa, bsp, info_):
    intervals = _get_info_intervals(info_)
    wa = _merge(wa, intervals)
    bsp = _merge(bsp, intervals)
    return wa, bsp


def _get_convex_hull(slice_, info_):
    ws, wa, bsp = slice_
    wa_rad = np.deg2rad(wa)
    points = np.column_stack([bsp * np.cos(wa_rad), bsp * np.sin(wa_rad)])
    try:
        vertices = ConvexHull(points).vertices
    except (ValueError, QhullError):
        return ws, wa, bsp, info_
    slice_ = slice_.T[vertices]
    if info_ is not None:
        info_ = [info_[vert] for vert in vertices]
        slice_, info_ = zip(*sorted(zip(slice_, info_), key=lambda x: x[0][1]))
        info_ = list(info_)
    else:
        slice_ = sorted(slice_, key=lambda x: x[1])
    slice_ = np.array(slice_).T

    if slice_[1, 0] == 0 and slice_[1, -1] == 360:
        ws, wa, bsp = slice_
        return ws, wa, bsp, info_

    if slice_[1][-1] - slice_[1][0] < 180:
        ws, wa, bsp = slice_
        return ws, wa, bsp, info_

    if slice_[1, 0] == 0:
        slice_ = np.column_stack([slice_, [slice_[0, 0], 360, slice_[2, 0]]])
        if info_ is not None:
            info_ = info_ + [info_[0]]
        ws, wa, bsp = slice_
        return ws, wa, bsp, info_

    if slice_[1, -1] == 360:
        slice_ = np.column_stack([[slice_[0, -1], 0, slice_[2, -1]], slice_])
        if info_ is not None:
            info_ = [info_[-1]] + info_
        ws, wa, bsp = slice_
        return ws, wa, bsp, info_

    # if wind angle difference is big, wrap around
    # estimate bsp value at 0 (360 resp)

    x0 = slice_[2, 0] * np.sin(np.deg2rad(slice_[1, 0]))
    x1 = slice_[2, -1] * np.sin(np.deg2rad(slice_[1, -1]))
    y0 = slice_[2, 0] * np.cos(np.deg2rad(slice_[1, 0]))
    y1 = slice_[2, -1] * np.cos(np.deg2rad(slice_[1, -1]))
    lamb = x0 / (x0 - x1)
    approx_ws = lamb * slice_[0, 0] + (1 - lamb) * slice_[0, -1]
    approx_bsp = lamb * y0 + (1 - lamb) * y1

    slice_ = np.column_stack(
        [[approx_ws, 0, approx_bsp], slice_, [approx_ws, 360, approx_bsp]]
    )
    if info_ is not None:
        info_ = [None] + info_ + [None]

    ws, wa, bsp = slice_

    return ws, wa, bsp, info_


def plot_polar(*args, **kwargs):
    """
    Creates a single `HROPolar` Axes and calls its `plot` method.
    Useful for simple plots of just one polar diagram.

    See also
    ----------
    `HROPolar.plot`
    """
    ax = plt.subplot(projection="hro polar")
    ax.plot(*args, **kwargs)


def scatter_polar(*args, **kwargs):
    """
    Creates a single `HROPolar` Axes and calls its `scatter` method.
    Useful for simple plots of just one polar diagram.

    See also
    ----------
    `HROPolar.scatter`
    """
    ax = plt.subplot(projection="hro polar")
    ax.scatter(*args, **kwargs)


def plot_flat(*args, **kwargs):
    """
    Creates a single `HROFlat` Axes and calls its `plot` method.
    Useful for simple plots of just one polar diagram.

    See also
    ----------
    `HROFlat.plot`
    """
    ax = plt.subplot(projection="hro flat")
    ax.plot(*args, **kwargs)


def scatter_flat(*args, **kwargs):
    """
    Creates a single `HROFlat` Axes and calls its `scatter` method.
    Useful for simple plots of just one polar diagram.

    See also
    ----------
    `HROFlat.scatter`
    """
    ax = plt.subplot(projection="hro flat")
    ax.scatter(*args, **kwargs)


def plot_color_gradient(*args, **kwargs):
    """
    Creates a single `HROColorGradient` Axes and calls its `plot` method.
    Useful for simple plots of just one polar diagram.

    See also
    ----------
    `HROColorGradient.plot`
    """
    ax = plt.subplot(projection="hro color gradient")
    ax.plot(*args, **kwargs)


def plot_3d(*args, **kwargs):
    """
    Creates a single `Axes3D` Axes and calls its `plot` method.
    Useful for simple plots of just one polar diagram.

    See also
    ----------
    `Axes3D.plot`
    """
    ax = plt.subplot(projection="hro 3d")
    ax.plot(*args, **kwargs)


def scatter_3d(*args, **kwargs):
    """
    Creates a single `Axes3D` Axes and calls its `scatter` method.
    Useful for simple plots of just one polar diagram.

    See also
    ----------
    `Axes3D.scatter`
    """
    ax = plt.subplot(projection="hro 3d")
    ax.scatter(*args, **kwargs)


def _configure_axes(ax, labels, colors, show_legend, legend_kw, **kwargs):
    _configure_colors(ax, labels, colors)
    _check_plot_kw(kwargs, True)
    if show_legend:
        if legend_kw is None:
            legend_kw = {}
        _configure_legend(ax, labels, colors, "True Wind Speed", **legend_kw)


def _set_polar_axis(ax):
    ax.set_theta_zero_location("N")
    ax.set_theta_direction("clockwise")


def _check_plot_kw(plot_kw, lines=True):
    ls = plot_kw.pop("linestyle", None) or plot_kw.pop("ls", None)
    if ls is None:
        plot_kw["ls"] = "-" if lines else ""
    else:
        plot_kw["ls"] = ls

    if plot_kw.get("marker", None) is None and not lines:
        plot_kw["marker"] = "o"


def _configure_colors(ax, ws, colors):
    if _only_one_color(colors):
        ax.set_prop_cycle("color", [colors])
        return

    if _more_colors_than_plots(ws, colors) or _no_color_gradient(colors):
        _set_color_cycle(ax, ws, colors)
        return

    _set_color_gradient(ax, ws, colors)


def _only_one_color(colors):
    return is_color_like(colors)


def _more_colors_than_plots(ws, colors):
    return len(ws) <= len(colors)


def _no_color_gradient(colors):
    all_color_format = all(_has_color_format(c) for c in colors)
    return len(colors) != 2 or not all_color_format


def _has_color_format(obj):
    if isinstance(obj, str):
        return True
    if len(obj) in [3, 4]:
        return True
    return False


def _set_color_cycle(ax, ws, colors):
    color_cycle = ["blue"] * len(ws)
    _configure_color_cycle(color_cycle, colors, ws)

    ax.set_prop_cycle("color", color_cycle)


def _configure_color_cycle(color_cycle, colors, ws):
    if isinstance(colors[0], tuple):
        for w, color in colors:
            i = list(ws).index(w)
            color_cycle[i] = color

        return

    colors = itertools.islice(colors, len(color_cycle))

    for i, color in enumerate(colors):
        color_cycle[i] = color


def _set_color_gradient(ax, ws, colors):
    color_gradient = _determine_color_gradient(colors, ws)
    ax.set_prop_cycle("color", color_gradient)


def _determine_color_gradient(colors, gradient):
    gradient_coeffs = _get_gradient_coefficients(gradient)
    color_gradient = _determine_colors_from_coefficients(
        gradient_coeffs, colors
    )
    return color_gradient


def _get_gradient_coefficients(gradient):
    gradient = np.asarray(gradient)
    min_gradient = gradient.min()
    max_gradient = gradient.max()

    return [
        (grad - min_gradient) / (max_gradient - min_gradient)
        for grad in gradient
    ]


def _determine_colors_from_coefficients(coefficients, colors):
    min_color = np.array(to_rgb(colors[0]))
    max_color = np.array(to_rgb(colors[1]))

    return [
        (1 - coeff) * min_color + coeff * max_color for coeff in coefficients
    ]


def _configure_legend(ax, ws, colors, label, **legend_kw):
    if _plot_with_color_gradient(ws, colors):
        _set_colormap(ws, colors, ax, label, **legend_kw)
        return

    if isinstance(colors[0], tuple) and not is_color_like(colors[0]):
        _set_legend_without_wind_speeds(ax, colors, legend_kw)
        return

    _set_legend_with_wind_speeds(ax, colors, ws, legend_kw)


def _plot_with_color_gradient(ws, colors):
    return not _no_color_gradient(colors) and len(ws) > len(colors) == 2


def _set_colormap(ws, colors, ax, label, **legend_kw):
    color_map = _create_color_map(colors)

    label_kw, legend_kw = _extract_possible_text_kw(legend_kw)
    plt.colorbar(
        ScalarMappable(
            norm=Normalize(vmin=min(ws), vmax=max(ws)), cmap=color_map
        ),
        ax=ax,
        **legend_kw,
    ).set_label(label, **label_kw)


def _extract_possible_text_kw(legend_kw):
    return {}, legend_kw


def _set_legend_without_wind_speeds(ax, colors, legend_kw):
    ax.legend(
        handles=[
            Line2D([0], [0], color=color, lw=1, label=f"TWS {ws}")
            for (ws, color) in colors
        ],
        **legend_kw,
    )


def _set_legend_with_wind_speeds(ax, colors, ws, legend_kw):
    slices = zip(ws, colors)

    ax.legend(
        handles=[
            Line2D([0], [0], color=color, lw=1, label=f"TWS {ws}")
            for (ws, color) in slices
        ],
        **legend_kw,
    )


def _set_3d_axis_labels(ax):
    ax.set_zlabel("TWS")
    ax.set_xlabel("Polar plane: TWA / BSP ")


def _remove_3d_tick_labels_for_polar_coordinates(ax):
    ax.xaxis.set_ticklabels([])
    ax.yaxis.set_ticklabels([])


def _create_color_map(colors):
    return LinearSegmentedColormap.from_list("cmap", list(colors))



"""
High level models regarding weather, influences to vessel performance and
latitude longitude projections.
"""

from .globe_model import FlatMercatorProjection, GlobeModel, SphericalGlobe
from .influencemodel import (
    IdentityInfluenceModel,
    InfluenceModel,
    WindAngleCorrectingInfluenceModel,
)
from .weather_model import (
    GriddedWeatherModel,
    MultiWeatherModel,
    NetCDFWeatherModel,
    WeatherModel,
)

__all__ = [
    "GlobeModel",
    "FlatMercatorProjection",
    "SphericalGlobe",
    "InfluenceModel",
    "IdentityInfluenceModel",
    "WindAngleCorrectingInfluenceModel",
    "WeatherModel",
    "GriddedWeatherModel",
    "NetCDFWeatherModel",
    "MultiWeatherModel",
]



"""
Contains the abstract base class `QualityAssurance` and some ready-to-use implementations aimed to measure the quality
of a polar diagram against test data.
"""

from abc import ABC, abstractmethod

import numpy as np

from hrosailing.core.computing import safe_operation


class QualityAssurance(ABC):
    """
    Base class for all quality assurance classes.
    """

    @abstractmethod
    def check(self, polar_diagram, test_data):
        """
        Method that should test if a given polar diagram works with given
        preprocessed test data and returns a dict containing relevant
        statistics.

        Parameters
        ----------
        polar_diagram : PolarDiagram

        test_data : (n, 3)-array
            The data to be tested against with columns referring to true wind speed, true wind angle and boat speed.

        Returns
        --------
        metrics : dict
            A dictionary containing relevant quality metrics.
        """


class MinimalQualityAssurance(QualityAssurance):
    """
    Quality assurance which evaluates the quality of a polar diagram by using a small set of well established
    statistical metrics.

    See also
    ----------
    `QualityAssurance`
    """

    def check(self, polar_diagram, test_data):
        """
        Returns
        -------
        statistics: dict
            Dictionary containing the following keys:

            - `"max_error"` : the maximal absolute difference between the boat speed of a test point the corresponding polar diagram output,
            - `"min_error"` : the minimal absolute difference between the boat speed of a test point the corresponding polar diagram output,
            - `"average_error"` : the average absolute difference between the boat speed of a test point the corresponding polar diagram output,
            - `"average_quadratic_error"` : the average quadratic difference between the boat speed of a test point the corresponding polar diagram output.

        See also
        ---------
        `QualityAssurance.check`
        """
        diffs = [abs(bsp - polar_diagram(ws, wa)) for ws, wa, bsp in test_data]
        statistics = {
            "max_error": max(diffs),
            "min_error": min(diffs),
            "average_error": np.mean(diffs),
            "average_quadratic_error": np.mean([diff**2 for diff in diffs]),
        }
        return statistics


class ComformingQualityAssurance(QualityAssurance):
    """
    Quality assurance which evaluates the quality of a polar diagram by using a small set of well established
    statistical metrics and additionally quality features derived from the theory of polar diagrams.

    See also
    ----------
    `QualityAssurance`
    """

    def check(self, polar_diagram, test_data):
        """
        Returns
        -------
        statistics: dict
            Dictionary containing the following keys:

            - `"max_error"` : the maximal absolute difference between the boat speed of a test point the corresponding polar diagram output,
            - `"min_error"` : the minimal absolute difference between the boat speed of a test point the corresponding polar diagram output,
            - `"average_error"` : the average absolute difference between the boat speed of a test point the corresponding polar diagram output,
            - `"average_quadratic_error"` : the average quadratic difference between the boat speed of a test point the corresponding polar diagram output,
            - `"max_zero_val"` : the maximal value of the polar diagram for wind angles of 0 or 360 degree and wind speed between 0 and 20,
            - `"min_zero_val"`: the minimal value of the polar diagram for wind angles of 0 or 360 degree and wind speed between 0 and 20,
            - `"average_zero_val"`: the average value of the polar diagram for wind angles of 0 or 360 degree and wind speed between 0 and 20,
            - `"average_quadratic_zero_val"`: the average squared value of the polar diagram for wind angles of 0 or 360 degree and wind speed between 0 and 20,
            - `"test_covering"`: number of unique test cases when rounded to the nearest integer,
            - `"local_test_data_difference"`: maximal difference in boat speeds in test cases where wind speed and wind angle are rounded to the same nearest integer respectively.

        See also
        ---------
        `QualityAssurance.check`
        """
        diffs = [abs(bsp - polar_diagram(ws, wa)) for ws, wa, bsp in test_data]
        zero_diffs = [
            abs(polar_diagram(ws, wa))
            for ws in np.linspace(0, 20, 20)
            for wa in [0, 360]
        ]
        tested_tuples = {}
        for ws, wa, bsp in test_data:
            test_tuple = round(ws), round(wa)
            if test_tuple not in tested_tuples:
                tested_tuples[test_tuple] = (np.inf, 0)
            prev_min, prev_max = tested_tuples[test_tuple]
            tested_tuples[test_tuple] = min(bsp, prev_min), max(bsp, prev_max)

        statistics = {
            "max_error": safe_operation(max, diffs),
            "min_error": safe_operation(min, diffs),
            "average_error": safe_operation(np.mean, diffs),
            "average_quadratic_error": safe_operation(
                np.mean, [diff**2 for diff in diffs]
            ),
            "max_zero_val": safe_operation(max, zero_diffs),
            "min_zero_val": safe_operation(min, zero_diffs),
            "average_zero_val": safe_operation(np.mean, zero_diffs),
            "average_quadratic_zero_val": safe_operation(
                np.mean, [diff**2 for diff in zero_diffs]
            ),
            "test_covering": len(tested_tuples),
            "average_local_test_data_difference": safe_operation(
                np.mean,
                [
                    bsp_max - bsp_min
                    for bsp_min, bsp_max in tested_tuples.values()
                ],
            ),
        }
        return statistics



"""
Classes used for modular modelling of different filtering methods based
on weights.

Defines the `Filter` abstract base class that can be used to create
custom filtering methods.

Subclasses of `Filter` can be used with the `PolarPipeline` class
in the `hrosailing.pipeline` module.
"""


from abc import ABC, abstractmethod

import numpy as np

from hrosailing.core.statistics import ComponentWithStatistics


class Filter(ComponentWithStatistics, ABC):
    """Base class for all filter classes."""

    @abstractmethod
    def filter(self, wts):
        """This method should be used, given an array of weights,
        to filter out points based on their weights, and produce a
        boolean array of the same size as `wts` and a dictionary containing
        statistics.

        Parameters
        ----------
        wts : numpy.ndarray of shape (n,)
            Weights of the points that are to be filtered, given
            as a sequence of scalars.

        Returns
        -------
        filtered_points : numpy.ndarray of shape (n,)
            Boolean array describing which points are filtered.
        """

    def set_statistics(self, filtered_points):
        """
        Computes standard statistics for the output of a filter.

        Parameters
        ----------
        filtered_points : boolean array
            Boolean array describing which points are filtered.
        """
        n_filtered_points = len([f for f in filtered_points if not f])

        super().set_statistics(
            n_filtered_points=n_filtered_points,
            n_rows=len(filtered_points) - n_filtered_points,
            error=None,
        )


class QuantileFilter(Filter):
    """A filter that removes the points with the lowest weights until as many points as
    possible are removed such that the total number of removed points is less than a given
    percentage.
    Supports the `repr` method.

    Parameters
    ----------
    percent : int or float, optional

        Defaults to `50`.

    See also
    ----------
    `Filter`
    """

    def __init__(self, percent=50):
        super().__init__()
        if percent < 0 or percent > 100:
            raise ValueError("`percent` is not between 0 and 100")

        self._percent = percent

    def __repr__(self):
        return f"QuantileFilter(percent={self._percent})"

    def filter(self, wts):
        """Filters a set of points given by their resp. weights
        according to the above described method.

        See also
        --------
        `Filter.filter`
        """
        if len(wts) == 0:
            return []
        filtered_points = self._calculate_quantile(wts)

        self.set_statistics(filtered_points)

        return filtered_points

    def _calculate_quantile(self, wts):
        return wts >= np.percentile(wts, self._percent)


class BoundFilter(Filter):
    """A filter that filters all points based on if their
    weight is outside an interval given by a lower and an upper bound.
    Supports the `repr` method.

    Parameters
    ----------
    lower_bound : int or float, optional
        The lower bound for the filter.

        Defaults to `0.5`.

    upper_bound : int or float, optional
        The upper bound for the filter.

        Defaults to `1`.

    See also
    ----------
    `Filter`
    """

    def __init__(self, lower_bound=0.5, upper_bound=1):
        super().__init__()
        if upper_bound < lower_bound:
            raise ValueError("`upper_bound` is smaller than `lower_bound`")

        self._u_b = upper_bound
        self._l_b = lower_bound

    def __repr__(self):
        return f"BoundFilter(upper_bound={self._u_b}, lower_bound={self._l_b})"

    def filter(self, wts):
        """Filters a set of points given by their resp. weights
        according to the above described procedure.

        See also
        --------
        `Filter.filter`
        """
        filtered_points = self._determine_points_within_bound(wts)

        self.set_statistics(filtered_points)

        return filtered_points

    def _determine_points_within_bound(self, wts):
        points_above_lower_bound = wts >= self._l_b
        points_below_upper_bound = wts <= self._u_b

        return points_above_lower_bound & points_below_upper_bound



"""
Contains the baseclass `Weigher` used in the `PolarPipeline` class
that can also be used to create custom weighers.

Also contains two predefined and usable weighers, the `CylindricMeanWeigher`
and the `CylindricMemberWeigher`, as well as the `WeightedPoints` class, used to
represent data points together with their respective weights.
"""


from abc import ABC, abstractmethod
from datetime import timedelta

import numpy as np

from hrosailing.core.computing import (
    data_dict_to_numpy,
    euclidean_norm,
    safe_operation,
    scaled_norm,
)
from hrosailing.core.constants import NORM_SCALES
from hrosailing.core.data import Data
from hrosailing.core.statistics import ComponentWithStatistics


class Weigher(ComponentWithStatistics, ABC):
    """Base class for all weigher classes.
    Basic arithmetic operations may be performed among weighers.
    """

    @abstractmethod
    def weigh(self, points) -> (np.ndarray, dict):
        """This method should be used to determine a weight for each point
        where the points might be given as `Data` or as `numpy.ndarray` of floating point numbers storing records
        in a row wise fashion and return the resulting weights.

        The following arithmetic operations are supported for `Weigher` instances `weigher`, `weigher1`, `weigher2`,
         integers `k` and `a` of type `float`, `int` or array like over `float` or `int`:

        - `weigher1 + weigher2`, weigher + a`
        - `-weigher`, `weigher1 - weigher2`, `weigher - a`
        - `a*weigher`, `weigher1*weigher2`
        - `weigher1/weigher2`, `a/weigher`, `weigher/a`
        - `weigher**k`

        The result will be a weigher producing weights according to the used formula.

        Parameters
        ----------
        points : Data or numpy.ndarray
        """

    def __add__(self, other):
        if isinstance(other, Weigher):
            sum_ = _BinaryMapWeigher(self, other, lambda x, y: x + y)
        elif isinstance(other, (int, float, np.ndarray)):
            sum_ = _UnaryMapWeigher(self, lambda x: x + other)
        return sum_

    def __radd__(self, other):
        return self + other

    def __mul__(self, other):
        if isinstance(other, Weigher):
            prod_ = _BinaryMapWeigher(self, other, lambda x, y: x * y)
        elif isinstance(other, (int, float, np.ndarray)):
            prod_ = _UnaryMapWeigher(self, lambda x: x * other)
        return prod_

    def __rmul__(self, other):
        return self * other

    def __sub__(self, other):
        if isinstance(other, Weigher):
            sub = _BinaryMapWeigher(self, other, lambda x, y: x - y)
        elif isinstance(other, np.ndarray):
            sub = _UnaryMapWeigher(self, lambda x: x - other)
        return sub

    def __neg__(self):
        neg = _UnaryMapWeigher(self, lambda x: -x)
        return neg

    def __truediv__(self, other):
        if isinstance(other, Weigher):
            div = _BinaryMapWeigher(self, other, lambda x, y: x / y)
        elif isinstance(other, (int, float, np.ndarray)):
            div = _UnaryMapWeigher(self, lambda x: x / other)
        return div

    def __rtruediv__(self, other):
        if isinstance(other, Weigher):
            div = _BinaryMapWeigher(self, other, lambda x, y: y / x)
        elif isinstance(other, np.ndarray):
            div = _UnaryMapWeigher(self, lambda x: other / x)
        return div

    def __pow__(self, power, modulo=None):
        pow_ = _UnaryMapWeigher(self, lambda x: x**power)
        return pow_

    def set_statistics(self, weights):
        """
        Produces hrosailing standard statistics for weighers, namely the average,
        minimal and maximal weight as well as a list of percentages of how many
        weights are contained in each 10th of the span between the minimal and
        maximal weight.
        The respective keys are `"average_weight"`, `"minimal_weight"`, `"maximal_weight"` and
        `"quantiles"`.

        Parameters
        ----------
        weights : array_like of floats
            The weights to be analyzed.
        """
        minw = safe_operation(np.min, weights)
        maxw = safe_operation(np.max, weights)
        span = safe_operation(lambda x: x[1] - x[0], [minw, maxw])

        def get_quantiles(args):
            minw, span = args
            return [
                round(
                    100
                    * len(
                        [
                            w
                            for w in weights
                            if (
                                minw + span * (i + 1) / 10
                                >= w
                                >= minw + span * i / 10
                            )
                        ]
                    )
                    / len(weights),
                    2,
                )
                for i in range(10)
            ]

        super().set_statistics(
            average_weight=safe_operation(np.mean, weights),
            minimal_weight=minw,
            maximal_weight=maxw,
            quantiles=safe_operation(get_quantiles, [minw, span]),
        )


class _UnaryMapWeigher(Weigher):
    def __init__(self, sub_weigher, map_):
        super().__init__()
        self._sub_weigher = sub_weigher
        self._map = map_

    def weigh(self, points) -> (np.ndarray, dict):
        weights, _ = self._sub_weigher.weigh(points)
        self.set_statistics(weights)
        return self._map(weights)


class _BinaryMapWeigher(Weigher):
    def __init__(self, sub_weigher, other_sub_weigher, map_):
        super().__init__()
        self._sub_weigher = sub_weigher
        self._other_sub_weigher = other_sub_weigher
        self._map = map_

    def weigh(self, points) -> (np.ndarray, dict):
        weights, _ = self._sub_weigher.weigh(points)
        other_weights, _ = self._other_sub_weigher.weigh(points)
        new_weights = self._map(weights, other_weights)
        self.set_statistics(new_weights)
        return new_weights


class AllOneWeigher(Weigher):
    """A weigher that functions mainly as a stand in if the weighing options
    of the pipeline are set to `False`. Weighs everything as `1`.

    See also
    ----------
    `Weigher`
    """

    def weigh(self, points) -> (np.ndarray, dict):
        """
        Assigns weights according to the procedure described above.

        See also
        --------
        `Weigher.weigh`
        """
        if isinstance(points, np.ndarray):
            size = len(points)
        else:
            size = points.n_rows
        return np.ones(size)


class CylindricMeanWeigher(Weigher):
    """A weigher that weighs given points according to the
    following procedure:

    For a given point `p` and points `pts` we look at all the points
    `pt` in `pts` such that :math:`||pt[:d-1] - p[:d-1]|| \\leq r`.

    Then we take the mean `m_p` and standard deviation `std_p`
    of the d-th component of all those points and set
    :math:`w_p = | m_p - p[d-1] | / std_p`.

    This weigher can only be used on floating point data.

    Parameters
    ----------
    radius : positive int or float, optional
        The radius of the considered cylinder, with infinite height, i.e. :math:`r`.

        Defaults to `0.05`.

    norm : function or callable, optional
        Norm with which to evaluate the distances, i.e. :math:`||.||`.

        If nothing is passed, it will automatically detect a scaled euclidean norm with respect to the used attributes.

    dimensions : list of str, optional
        If not `None` and the input of `weigh` is of type `Data`, only the attributes mentioned in
        `attributes` will be considered.

        Defaults to `None`.

    See also
    ----------
    `Weigher`
    """

    def __init__(self, radius=0.05, norm=None, dimensions=None):
        super().__init__()
        if radius <= 0:
            raise ValueError(
                f"Invalid value for `radius`: {radius}. Non-positive number."
                " Use a positive value for `radius`."
            )

        self._radius = radius

        self._norm = norm
        self._dimensions = dimensions

    def __repr__(self):
        return (
            f"CylindricMeanWeigher(radius={self._radius}, "
            f"norm={self._norm.__name__})"
        )

    def weigh(self, points):
        """Weighs given points according to the procedure described above.

        Parameters
        ----------
        points : numpy.ndarray of shape (n, d) with d>=3 or Data
            Points to be weighed. Data has to be given in type `float`.

        Returns
        -------
        WeightedPoints : numpy.ndarray of shape (n,)
            Normalized weights of the input points.

        See also
        ----------
        `Weigher.weigh`
        """
        if isinstance(points, np.ndarray) and len(points) == 0:
            return []
        if isinstance(points, Data) and points.n_rows == 0:
            return []
        dimensions, points, bsps = _set_points_from_data(
            points, self._dimensions
        )
        weights = [
            self._calculate_weight(point, points, bsps, bsp, dimensions)
            for point, bsp in zip(points, bsps)
        ]
        weights = np.array(weights)
        weights = 1 - _normalize(weights, np.max)

        self.set_statistics(weights)

        return weights

    def _calculate_weight(self, point, points, bsps, bsp, dimensions):
        points_in_cylinder = self._determine_points_in_cylinder(
            point, points, bsps, dimensions
        )

        std = _standard_deviation_of(points_in_cylinder)
        mean = _mean_of(points_in_cylinder)

        return np.abs(mean - bsp) / std

    def _determine_points_in_cylinder(self, point, points, bsps, dimensions):
        if self._norm is None:
            self._norm = hrosailing_standard_scaled_euclidean_norm(dimensions)
        in_cylinder = self._norm(points - point) <= self._radius
        return bsps[in_cylinder]


def _standard_deviation_of(points_in_cylinder):
    return np.std(points_in_cylinder) or 1  # in case that there are no points


def _mean_of(points_in_cylinder):
    return np.mean(points_in_cylinder)


def _normalize(weights, normalizer):
    if len(weights) == 0:
        return weights
    normalizer_ = normalizer(weights)
    if normalizer_ == 0:
        return weights
    return weights / normalizer(weights)


class CylindricMemberWeigher(Weigher):
    """A weigher that weighs given points according to the
    following procedure:

    For a given point `p` and points `pts`
    we look at all the points `pt` in `pts` such that
    :math:`|pt[-1] - p[-1]| \\leq h` and :math:`||pt[:d-1] - p[:d-1]|| \\leq r`.

    If we call the set of all such points `P`, then :math:`w_p = |P| - 1`
    normalized by dividing by the maximum weight.

    Parameters
    ----------
    radius : positive int or float, optional
        The radius of the considered cylinder, i.e. :math:`r`.

        Defaults to `0.05`

    length : non-negative int or float, optional
        The height of the considered cylinder, i.e. :math:`h`.

        If length is 0, the cylinder is a d-1 dimensional ball.

        Defaults to `0.05`.

    norm : function or callable, optional
        Norm with which to evaluate the distances, i.e. :math:`||.||`.

        If nothing is passed, it will default to :math:`||.||_2`.

    dimensions : [str] or None, optional
        If the data is given as `Data`, `attributes` contains the keys
        which will be used in order to create the data space.
        If `None`, all keys of the given `Data` are used.

        Defaults to `None`.

    See also
    ----------
    `Weigher`
    """

    def __init__(self, radius=0.05, length=0.05, norm=None, dimensions=None):
        super().__init__()
        if radius <= 0:
            raise ValueError("`radius` is non-positive")

        if length < 0:
            raise ValueError("`length` should be non-negative.")

        self._radius = radius
        self._length = length

        if norm is None:
            self._norm = hrosailing_standard_scaled_euclidean_norm(dimensions)
        else:
            self._norm = norm
        self._dimensions = dimensions

    def __repr__(self):
        return (
            f"CylindricMemberWeigher(radius={self._radius},"
            f"length={self._length}, norm={self._norm.__name__})"
        )

    def weigh(self, points):
        """Weighs given points according to the procedure described above.

        Parameters
        ----------
        points : numpy.ndarray of shape (n, d)
            Points to be weighed.

        Returns
        -------
        weights : numpy.ndarray of shape (n,)
            Normalized weights of the input points.

        See also
        ----------
        `Weigher.weigh`
        """

        if isinstance(points, np.ndarray) and len(points) == 0:
            return []
        if isinstance(points, Data) and points.n_rows == 0:
            return []

        self._dimensions, points = _set_points_from_data(
            points, self._dimensions, reduce=False
        )

        weights = [self._calculate_weight(point, points) for point in points]
        weights = np.array(weights)
        weights = _normalize(weights, np.max)

        self.set_statistics(weights)
        return weights

    def _calculate_weight(self, point, points):
        n_points_in_cylinder = self._count_points_in_cylinder(point, points)
        return n_points_in_cylinder - 1

    def _count_points_in_cylinder(self, point, points):
        height = np.abs(points[:, -1] - point[-1]) <= self._length
        radius = self._norm(points[:, :-1] - point[:-1]) <= self._radius

        cylinder = height & radius
        points_in_cylinder = cylinder[cylinder]
        return len(points_in_cylinder)


class FluctuationWeigher(Weigher):
    """
    A weigher that benefits data with (locally) small fluctuation.

    If a single attribute is used, we use the following procedure:

    For each data point, the weigher considers the standard deviation of the data restricted to an interval
    before and after the datestamp of the considered point.
    Given an upper bound on the standard deviation the weight is computed by a ReLu function mirrored and stretched
    such that a standard deviation of 0 gets the weight 1 and if the standard variation exceeds the upper bound, the
    weight is set to 0.

    If more than one attribute is used, the weights described above will be computed for each such attribute
    and will then be multiplied.

    Parameters
    ----------
    attributes : list of str
        The names of the numeric attributes to be considered.

    timespan : timedelta or tuple of two timedelta

        - If it is a single value, the weigher will consider the corresponding duration before the examined
        data point,
        - If it is a tuple, the weigher will consider points in the interval
        (`time` - `timespan[0]`, `time` + `timespan[1]`) where `time` is the timestamp of the currently examined data
        point.

    upper_bounds : list of int or float
        The upper bounds on the standard deviation for each considered attribute.

    See also
    ----------
    `Weigher`
    """

    def __init__(self, attributes, timespan, upper_bounds):
        super().__init__()
        if isinstance(timespan, timedelta):
            self._timespan_before = timespan
            self._timespan_after = timedelta(seconds=0)
        else:
            self._timespan_before = timespan[0]
            self._timespan_after = timespan[1]
        self._attributes = attributes
        self._upper_bounds = np.array(upper_bounds)

    def weigh(self, points):
        """
        Weighs points by the procedure described above.

        Parameters
        ----------
        points : Data
            Has to contain the key `"datetime"` and all keys contained in the `dimension` parameter during
            initialization. Fields which do not contain `float` values are ignored.

        See also
        --------
        `Weigher.weigh`
        """
        times = points["datetime"]
        dimensions, points = _set_points_from_data(
            points, self._attributes, False
        )
        upper_bounds = self._upper_bounds[
            [i for i, key in enumerate(self._attributes) if key in dimensions]
        ]

        weights = [1] * len(points)

        for curr_idx, (dt, pt) in enumerate(zip(times, points)):
            start_idx = min(
                i
                for i in range(len(points))
                if times[i] <= dt and dt - times[i] <= self._timespan_before
            )
            end_idx = max(
                i
                for i in range(len(points))
                if dt <= times[i] and times[i] - dt <= self._timespan_after
            )
            for col, ub in enumerate(upper_bounds):
                curr_pts = points[start_idx : end_idx + 1, col]
                std = np.std(curr_pts)
                if std > ub:
                    weights[curr_idx] = 0
                else:
                    weights[curr_idx] *= (ub - std) / ub

        self.set_statistics(weights)

        return weights


def _set_points_from_data(data, attributes, reduce=True):
    if isinstance(data, np.ndarray):
        if reduce:
            bsp = data[:, -1]
            data = data[:, :-1]
            return attributes, data, bsp
        return attributes, data

    if attributes is None:
        attributes = list(data.keys()) or list(data.keys)

    if reduce:
        if "BSP" in attributes:
            bsp = data["BSP"]
            attributes.remove("BSP")
        elif "SOG" in attributes:
            bsp = data["SOG"]
            attributes.remove("SOG")
        else:
            raise ValueError(
                "Data has to support one of the keys 'BSP' or 'SOG'"
            )

    if isinstance(data, dict):
        data = data_dict_to_numpy(data, attributes)

    if isinstance(data, Data):
        attributes, data = data[attributes].numerical

    if reduce:
        return attributes, data, np.asarray(bsp)
    return attributes, data


class FuzzyBool:
    """
    Class representing a fuzzy truth statement, i.e. a function with values
    between 0 and 1 (truth function).
    The truth function can be evaluated by calling the instance.
    Supports the operators `&`, `|`, `~` for the operators and, or, not.
    If `fuzzy_bool` is a `FuzzyBool` instance, then `fuzzy_bool[item]` will return a `FuzzyBool` instance
    evaluating the respective truth function on `x[item]` instead of `x` for a given input `x`.

    The easiest way to initialize a `FuzzyBool` object is by using the operators of the
    `FuzzyVariables` class, but it can also be initialized with a custom truth
    function.

    Also supports the `str` method.

    Parameters
    ----------
    eval_fun : callable
        The truth function used. Truth values between 0 and 1 are recommended.

    Attributes
    ----------
    repr : str
        A human readable representation of the `FuzzyBool` object.
        Is returned by `__str__`. First set to 'Fuzzy-Bool'.

    See also
    --------
    For recommendations on how to use a `FuzzyBool` see also `FuzzyVariable`.
    """

    def __init__(self, eval_fun):
        self._fun = eval_fun
        self.repr = "Fuzzy-Bool"

    def __call__(self, x):
        return self._fun(x)

    def __and__(self, other):
        new = FuzzyBool.fuzzy_and(self, other)
        new._concat_repr(self, other, "&")
        return new

    def __or__(self, other):
        new = FuzzyBool.fuzzy_or(self, other)
        new._concat_repr(self, other, "|")
        return new

    def __invert__(self):
        new = FuzzyBool.fuzzy_not(self)
        new.repr = f"~({self})"
        return new

    def __getitem__(self, item):
        def eval_fun(x):
            return self._fun(x[item])

        new = FuzzyBool(eval_fun)
        new.repr = f"{self}[{item}]"
        return new

    def __str__(self):
        return self.repr

    @classmethod
    def fuzzy_and(cls, one, other):
        """
        Parameters
        ----------
        one : FuzzyBool

        other : FuzzyBool

        Returns
        -------
        one_and_other : FuzzyBool
            Representation of the fuzzy 'and' operation of `one` and `other`
            realized via taking the minimum.
        """

        def eval_fun(x):
            concat = np.row_stack([one(x), other(x)])
            return np.min(concat, axis=0)[0]

        return cls(eval_fun)

    @classmethod
    def fuzzy_or(cls, one, other):
        """
        Parameters
        ----------
        one : FuzzyBool

        other : FuzzyBool

        Returns
        -------
        one_or_other : FuzzyBool
            Representation of the fuzzy 'or' operation of `one` and `other`
            realized via taking the maximum.
        """

        def eval_fun(x):
            concat = np.row_stack([one(x), other(x)])
            return np.max(concat, axis=0)[0]

        return cls(eval_fun)

    @classmethod
    def fuzzy_not(cls, one):
        """
        Parameters
        ----------
        one : FuzzyBool

        Returns
        -------
        not_one : FuzzyBool
            Representation of the fuzzy 'not' operation of `one`
            realized via taking the difference to 1.
        """
        return cls(lambda x: 1 - one(x))

    @classmethod
    def sigmoid(cls, center, sharpness, sigma):
        """
        Classical activation function.

        Parameters
        ----------
        center : int or float

        sharpness : int or float
            Controls the slope of the sigmoid function
            (higher sharpness yields higher slope).

        sigma : {1, -1}
            The direction of the sigmoid function; -1 yields the classical
            sigmoid, 1 yields the inverted sigmoid.

        Returns
        -------
        sigmoid : FuzzyBool
            A `FuzzyBool` object with truth function
            :math:`x → 1/(1+e^{sigma * sharpness * (x - center)})`.
        """

        def eval_fun(x):
            return 1 / (1 + np.exp(sigma * sharpness * (x - center)))

        new = cls(eval_fun)
        new.repr = (
            f"sigmoid(center={center},sharpness={sharpness},sigma={sigma})"
        )
        return new

    def _concat_repr(self, origin, other, concat):
        """
        Set the representation string to f"{origin} {concat} {other}". Use brackets if necessary.

        Parameters
        ----------
        origin : FuzzyBool

        other : FuzzyBool

        concat : str
        """
        self_repr = f"({origin})" if " " in f"{origin}" else f"{origin}"
        other_repr = f"({other})" if " " in f"{other}" else f"{other}"
        self.repr = f"{self_repr} {concat} {other_repr}"


class FuzzyVariable:
    """
    Refers to Variables in the fuzzy logic.
    It's main purpose is to easily create `FuzzyBool` instances.

    For example, the following notations work for a `FuzzyVariable` `x`, `int` or `float`
    Variables `a`, `b`, `s` and `key` such that `x.__getitem__(key)` works:

    - `x < a`, `x <= a`, `x > a`, `x >= a`, `x == a`
            refers to the respective truth function
            (using sigmoid activation function),

    - `x(s) < a`, ... (same as above, but with sharpness `s` used),

    - `x[key] < a`, ... (same as above, but the truth function will be
            applied after getting the item referenced by `key`),

    - `x[key](s) < a`, ... (the both notations above combined),

    - `(x < a) & (x > b)`, ... ('and' concatenation),

    - `(x < a) | (x > b)`, ... ('or' concatenation),

    - `~(x < a)`, ... ('not' operation).

    Also supports the methods `repr` and `str`.

    Parameters
    ----------
    key : None or str
        If `key` is not `None`, all generated `FuzzyBool` instances apply the
        truth function to `x`[`key`] instead of `x`.

        Defaults to `None`.

    sharpness : int
        Defines the default sharpness of all `FuzzyBool` instances generated
        by `FuzzyBool.sigmoid`.
        This sharpness will be used if no other sharpness is given via the
        `__call__`-method.

    See also
    --------
    `FuzzyBool`
    """

    def __init__(self, sharpness=10, key=None):
        self.key = key
        self._sharpness = sharpness
        self._next_sharpness = sharpness

    @property
    def sharpness(self):
        """Sharpness that will be used for the next generated `FuzzyBool` instance."""
        next_sharpness = self._next_sharpness
        self._next_sharpness = self._sharpness
        return next_sharpness

    def _truth(self, other, sigma):
        sigmoid = FuzzyBool.sigmoid(float(other), self.sharpness, sigma)
        if self.key is None:
            return sigmoid

        return sigmoid[self.key]

    def __gt__(self, other):
        new = self._truth(other, -1)
        new.repr = f"{self} > {other}"
        return new

    def __lt__(self, other):
        new = self._truth(other, 1)
        new.repr = f"{self} < {other}"
        return new

    def __ge__(self, other):
        return self > other

    def __le__(self, other):
        return self < other

    def __eq__(self, other):
        new = FuzzyBool.fuzzy_and(self < other, self > other)
        new.repr = f"{self} == {other}"
        return new

    def __getitem__(self, item):
        return FuzzyVariable(key=item, sharpness=self.sharpness)

    def __call__(self, sharpness):
        self._next_sharpness = sharpness
        return self

    def __str__(self):
        if self.key is None:
            return "x"
        return f"x[{self.key}]"

    def __repr__(self):
        if self.key is None:
            return f"x({self._sharpness})"
        return f"x({self._sharpness})[{self.key}]"


class FuzzyWeigher(Weigher):
    """
    Weigher that uses the truth function of a `FuzzyBool` object to create
    weights.

    Parameters
    ----------
    fuzzy : FuzzyBool
        The object wrapped around the truth function.

    See also
    ---------
    `FuzzyBool`, `FuzzyVariable`, `Weigher`
    """

    def __init__(self, fuzzy):
        super().__init__()
        self.fuzzy = fuzzy

    def weigh(self, points):
        """
        See also
        --------
        `Weigher.weigh`
        """
        if isinstance(points, Data):
            weights = np.array([self.fuzzy(point) for point in points.rows()])
        elif isinstance(points, np.ndarray):
            weights = np.array([self.fuzzy(point) for point in points])
        else:
            raise TypeError(
                "FuzzyWeigher only takes numpy arrays or"
                " `hrosailing.core.data.Data` objects, got"
                f" {type(points)} instead."
            )
        self.set_statistics(weights)
        return weights


def hrosailing_standard_scaled_euclidean_norm(dimensions=None):
    """
    Returns a scaled euclidean norm function where the scales are chosen with respect to `constants.NORM_SCALES`
    (or 1 if `constants.NORM_SCALES` does not contain the respective key).

    Parameters
    ----------
    dimensions : iterator of str or None, optional
        Iterates over the names of the attributes used.
        If set to `None`, a two-dimensional norm with scalings 1/20 (for wind speed) and 1/360 (for wind angle) is
        returned.

        Defaults to `None`.

    Returns
    ---------
    norm : callable
        The euclidean norm, scaled by coefficients corresponding to
        `constants.NORM_SCALES`.
    """
    if dimensions is None:
        dimensions = ["TWS", "TWA"]
    scales = [
        NORM_SCALES[key] if key in NORM_SCALES else 1 for key in dimensions
    ]
    return scaled_norm(euclidean_norm, scales)



"""
Methods for reading polar diagram information from files.
"""

import csv
from ast import literal_eval

import numpy as np

from hrosailing.core.exceptions import FileReadingException
from hrosailing.polardiagram._polardiagramtable import (
    PolarDiagram,
    PolarDiagramTable,
)


def from_csv(csv_path, fmt="hro"):
    """Reads a .csv file and returns the `PolarDiagram`
    instance contained in it.

    Parameters
    ----------
    csv_path : path-like
        Path to a .csv file.

    fmt : str
        The format of the .csv file.

        - `"hro"` : format created by the `to_csv`-method of the `PolarDiagram` class,
        - `"orc"` : format found at [ORC](https://\
            jieter.github.io/orc-data/site/),
        - `"opencpn"` : format created by the [OpenCPN Polar Plugin](https://\
            opencpn.org/OpenCPN/plugins/polar.html),
        - `"array"` : tab-separated polar diagram in form of a table, also
            see the example files for a better look at the format.

    Returns
    -------
    out : PolarDiagram
        `PolarDiagram` instance contained in the .csv file.

    Raises
    ------
    FileReadingException
        If, in the format `hro`, the first row does not match any
        `PolarDiagram` subclass.

    OSError
        If file does not exist or no read permission for that file is given.

    Examples
    --------
    (For the following and more files also see
    [examples](https://github.com/hrosailing/hrosailing/tree/main/examples))

        >>> from polardiagram._reading import from_csv
        >>> pd = from_csv("table_hro_format_example.csv", fmt="hro")
        >>> print(pd)
          TWA / TWS    6.0    8.0    10.0    12.0    14.0    16.0    20.0
        -----------  -----  -----  ------  ------  ------  ------  ------
        52.0          3.74   4.48    4.96    5.27    5.47    5.66    5.81
        60.0          3.98   4.73    5.18    5.44    5.67    5.94    6.17
        75.0          4.16   4.93    5.35    5.66    5.95    6.27    6.86
        90.0          4.35   5.19    5.64    6.09    6.49    6.70    7.35
        110.0         4.39   5.22    5.68    6.19    6.79    7.48    8.76
        120.0         4.23   5.11    5.58    6.06    6.62    7.32    9.74
        135.0         3.72   4.64    5.33    5.74    6.22    6.77    8.34
        150.0         3.21   4.10    4.87    5.40    5.78    6.22    7.32
    """
    if fmt not in {"array", "hro", "opencpn", "orc"}:
        raise ValueError("`fmt` unknown")

    with open(csv_path, "r", newline="", encoding="utf-8") as file:
        if fmt == "hro":
            return _read_intern_format(file)

        return _read_extern_format(file, fmt)


def _read_intern_format(file):
    subclasses = {cls.__name__: cls for cls in PolarDiagram.__subclasses__()}

    first_row = file.readline().rstrip()
    if first_row not in subclasses:
        raise FileReadingException(
            f"no polar diagram format with the name {first_row} exists"
        )

    pd = subclasses[first_row]
    return pd.__from_csv__(file)


def _read_extern_format(file, fmt):
    if fmt == "array":
        ws_res, wa_res, bsps = _read_from_array(file)
    elif fmt == "orc":
        ws_res, wa_res, bsps = _read_orc_format(file)
    else:
        ws_res, wa_res, bsps = _read_opencpn_format(file)

    return PolarDiagramTable(ws_res, wa_res, bsps)


def _read_from_array(file):
    file_data = np.genfromtxt(file)
    return file_data[0, 1:], file_data[1:, 0], file_data[1:, 1:]


def _read_orc_format(file):
    csv_reader = csv.reader(file, delimiter=";")

    ws_res = _read_wind_speeds(csv_reader)

    # skip line of zeros
    next(csv_reader)

    wa_res, bsps = _read_wind_angles_and_boat_speeds(csv_reader)

    return ws_res, wa_res, bsps


def _read_wind_speeds(csv_reader):
    return [literal_eval(ws) for ws in next(csv_reader)[1:]]


def _read_wind_angles_and_boat_speeds(csv_reader):
    wa_res = []
    bsps = []

    for row in csv_reader:
        wa_res.append(literal_eval(row[0].replace("°", "")))
        bsps.append([literal_eval(bsp) if bsp != "" else 0 for bsp in row[1:]])

    return wa_res, bsps


def _read_opencpn_format(file):
    csv_reader = csv.reader(file, delimiter=",")

    ws_res = _read_wind_speeds(csv_reader)
    wa_res, bsps = _read_wind_angles_and_boat_speeds(csv_reader)

    return ws_res, wa_res, bsps
